"""Defines all execution graph node types of expressions. An expression cannot
be executed alone and usually has no side-effects."""
from typing import Literal, TypedDict


LiteralValueType = str | int | float | bool | list | None
"""A literal value."""
JSONType = str | int | float | list | dict | None
"""A literal value that can be converted to JSON."""


ValueType = Literal["str", "int", "float", "bool", "list", "none"]
"""Named literal value types."""


BinOps = Literal[
    "add",
    "sub",
    "and",
    "or",
    "eq",
    "ne",
    "lt",
    "gt",
    "le",
    "ge",
]
"""Binary operations."""


ArgObj = TypedDict('ArgObj', {
    "kind": Literal["arg"],
    "name": str,
    "readable": str,
})
"""Reads a script argument."""
KeyObj = TypedDict('KeyObj', {
    "kind": Literal["key"],
    "name": str,
    "readable": str,
})
"""Reads a script key argument."""
VarObj = TypedDict('VarObj', {
    "kind": Literal["var"],
    "name": str,
})
"""Reads a local variable."""
IndexObj = TypedDict('IndexObj', {
    "kind": Literal["index"],
    "name": str,
})
"""Reads an index variable."""
RefIdObj = ArgObj | KeyObj | VarObj | IndexObj
"""References some variable."""


LoadArg = TypedDict('LoadArg', {
    "kind": Literal["load_json_arg", "load_key_arg"],
    "index": int,
})
"""Loads a (key or value) argument. Usually used to assign to an argument
variable."""
LiteralValObj = TypedDict('LiteralValObj', {
    "kind": Literal["val"],
    "value": LiteralValueType,
    "type": ValueType,
})
"""Loads a literal value."""
ConstantObj = TypedDict('ConstantObj', {
    "kind": Literal["constant"],
    "raw": str,
})
"""Loads a named constant."""
UnaryOpObj = TypedDict('UnaryOpObj', {
    "kind": Literal["unary"],
    "op": Literal["not"],
    "arg": 'ExprObj',
})
"""Performs a unary operation."""
BinaryOpObj = TypedDict('BinaryOpObj', {
    "kind": Literal["binary"],
    "op": BinOps,
    "left": 'ExprObj',
    "right": 'ExprObj',
})
"""Performs a binary operation."""
ArrayAtObj = TypedDict('ArrayAtObj', {
    "kind": Literal["array_at"],
    "var": RefIdObj,
    "index": 'ExprObj',
})
"""Reads an index from an array."""
ArrayLengthObj = TypedDict('ArrayLengthObj', {
    "kind": Literal["array_len"],
    "var": RefIdObj,
})
"""Reads the length of an array."""
ConcatObj = TypedDict('ConcatObj', {
    "kind": Literal["concat"],
    "strings": 'list[ExprObj]',
})
"""String concatenates a sequence of expressions. The expressions should be
strings."""
CallObj = TypedDict('CallObj', {
    "kind": Literal["call"],
    "name": str,
    "args": 'list[ExprObj]',
    "no_adjust": bool,
})
"""Calls a function."""
ExprObj = (
    RefIdObj
    | LoadArg
    | LiteralValObj
    | ConstantObj
    | UnaryOpObj
    | BinaryOpObj
    | ArrayAtObj
    | ArrayLengthObj
    | ConcatObj
    | CallObj
)
"""An expression evaluates to a value."""


def get_literal(obj: ExprObj, vtype: ValueType | None = None) -> JSONType:
    """
    Returns the literal value of a literal expression object.

    Args:
        obj (ExprObj): The literal expression.

        vtype (ValueType | None, optional): The expected type.
        Defaults to None.

    Returns:
        JSONType: The literal value if the expression is a literal and the
        value has the correct type.
    """
    if obj["kind"] != "val":
        return None
    if vtype is not None and obj["type"] != vtype:
        return None
    return obj["value"]


def is_none_literal(obj: ExprObj) -> bool:
    """
    Whether a literal expression is None.

    Args:
        obj (ExprObj): The literal expression.

    Returns:
        bool: Whether it contains None as literal.
    """
    if obj["kind"] != "val":
        return False
    if obj["type"] != "none":
        return False
    return True


def find_literal(
        objs: list[ExprObj],
        value: JSONType,
        *,
        vtype: ValueType | None = None,
        no_case: bool = False) -> tuple[int, JSONType] | None:
    """
    Searches for a specified literal in a list of expressions.

    Args:
        objs (list[ExprObj]): The list of expressions.

        value (JSONType): The value to look for.

        vtype (ValueType | None, optional): The type of the value to look for.
        Defaults to None.

        no_case (bool, optional): Whether the value should be search without
        considering case (only for strings). Defaults to False.

    Returns:
        tuple[int, JSONType] | None: If found, the index and the literal.
        Otherwise None.
    """
    if vtype != "none" and value is not None:

        def value_check(obj: ExprObj) -> tuple[bool, JSONType]:
            res = get_literal(obj, vtype)
            if no_case and vtype == "str":
                return (f"{res}".upper() == f"{value}".upper(), res)
            return (res == value, res)

        check = value_check
    else:

        def none_check(obj: ExprObj) -> tuple[bool, JSONType]:
            is_none = is_none_literal(obj)
            return (is_none, None)

        check = none_check

    for ix, obj in enumerate(objs):
        is_hit, val = check(obj)
        if is_hit:
            return (ix, val)
    return None



"""This module handles the internal state of the memory runtime."""
import collections
import datetime
import itertools
import time
from typing import Literal, overload

from redipy.api import RedisAPI, RSetMode, RSM_ALWAYS, RSM_EXISTS, RSM_MISSING
from redipy.util import now, time_diff, to_number_str


def compute_expire(
        *,
        expire_timestamp: datetime.datetime | None,
        expire_in: float | None) -> float | None:
    """
    Computes the monotonic time point when a key should expire.

    Args:
        expire_timestamp (datetime.datetime | None): An absolute timestamp.
        expire_in (float | None): A relative time difference in seconds.

    Raises:
        ValueError: If both expire_timestamp and expire_in are set.

    Returns:
        float | None: The monotonic time point.
    """
    if expire_timestamp is None:
        if expire_in is None:
            return None
        return time.monotonic() + expire_in
    if expire_in is not None:
        raise ValueError(
            f"cannot set timestamp {expire_timestamp} "
            f"and duration {expire_in} at the same time")
    return time.monotonic() + time_diff(now(), expire_timestamp)


KeyType = Literal[
    "value",
    "list",
    "hash",
    "zset",
]
"""The different key types."""


class State:
    """
    The state holding the actual values of the memory runtime. A state can have
    a parent state which values it shadows. The parent can be updated by
    applying all changes.
    """
    def __init__(self, parent: 'State | None' = None) -> None:
        """
        Creates a new state.

        Args:
            parent (State | None, optional): The optional parent whose values
            this state shadows. Defaults to None.
        """
        super().__init__()
        self._parent = parent
        self._vals: dict[str, tuple[str, float | None]] = {}
        self._queues: dict[str, collections.deque[str]] = {}
        self._hashes: dict[str, dict[str, str]] = {}
        # TODO: to be replaced by something more efficient later
        self._zorder: dict[str, list[str]] = {}
        self._zscores: dict[str, dict[str, float]] = {}
        self._deletes: set[str] = set()

    def key_type(self, key: str) -> KeyType | None:
        """
        Computes the type of a given key.

        Args:
            key (str): The key.

        Returns:
            KeyType | None: The type of the key or None if it doesn't exist.
        """
        if key in self._deletes:
            return None
        if key in self._vals:
            return "value"
        if key in self._queues:
            return "list"
        if key in self._hashes:
            return "hash"
        if key in self._zorder:
            return "zset"
        if self._parent is not None:
            return self._parent.key_type(key)
        return None

    def verify_key(self, key_type: KeyType, key: str) -> None:
        """
        Ensures that the given key does not already exist with a different
        type.

        Args:
            key_type (KeyType): The expected key type.
            key (str): The key.

        Raises:
            ValueError: If the key already exists with a different type.
        """
        if key in self._deletes:
            return
        if key_type != "value" and key in self._vals:
            raise ValueError(f"key {key} already used as value")
        if key_type != "list" and key in self._queues:
            raise ValueError(f"key {key} already used as list")
        if key_type != "hash" and key in self._hashes:
            raise ValueError(f"key {key} already used as hash")
        if key_type != "zset" and key in self._zorder:
            raise ValueError(f"key {key} already used as sorted set")
        if self._parent is not None:
            self._parent.verify_key(key_type, key)

    def exists(self, key: str) -> bool:
        """
        Checks whether a given key exists.

        Args:
            key (str): The key.

        Returns:
            bool: Whether it exists as any type.
        """
        if key in self._deletes:
            return False
        if key in self._vals:
            return True
        if key in self._queues:
            return True
        if key in self._hashes:
            return True
        if key in self._zorder:
            return True
        if self._parent is not None:
            return self._parent.exists(key)
        return False

    def delete(self, deletes: set[str]) -> None:
        """
        Deletes keys.

        Args:
            deletes (set[str]): The keys to delete.
        """
        if self._parent is not None:
            self._deletes.update(deletes)
        for key in deletes:
            self._vals.pop(key, None)
            self._queues.pop(key, None)
            self._hashes.pop(key, None)
            self._zorder.pop(key, None)
            self._zscores.pop(key, None)

    def apply(self, other: 'State') -> None:
        """
        Applies the given state. All updates that are different in the other
        state will be updated in the current state and all deleted keys in
        the other state will delete corresponding keys in the current state.
        If a key in the other state has a different type than the key in the
        current state the type in the current state will be changed. This
        relies on the invariant that a key can only appear once in the other
        and once in the current state.

        Args:
            other (State): The state to be applied to the current state.
        """
        raw_vals = other.raw_vals()
        raw_queues = other.raw_queues()
        raw_hashes = other.raw_hashes()
        raw_zorder = other.raw_zorder()
        raw_zscores = other.raw_zscores()

        new_keys: set[str] = set()  # NOTE: new means new for the given type
        new_keys.update(set(raw_vals.keys()) - set(self._vals.keys()))
        new_keys.update(set(raw_queues.keys()) - set(self._queues.keys()))
        new_keys.update(set(raw_hashes.keys()) - set(self._hashes.keys()))
        new_keys.update(set(raw_zorder.keys()) - set(self._zorder.keys()))
        new_keys.update(set(raw_zscores.keys()) - set(self._zscores.keys()))
        # NOTE: deleting all new keys makes sure they don't exist as a
        # different type
        self.delete(new_keys)

        self._vals.update(raw_vals)
        self._queues.update(raw_queues)
        self._hashes.update(raw_hashes)
        self._zorder.update(raw_zorder)
        self._zscores.update(raw_zscores)
        self._clean_vals()
        self.delete(other.raw_deletes())

    def reset(self) -> None:
        """
        Completely resets the state.
        """
        self._vals.clear()
        self._queues.clear()
        self._hashes.clear()
        self._zorder.clear()
        self._zscores.clear()
        self._deletes.clear()

    def _prepare_key_for_write(self, key: str) -> None:
        """
        Prepares a key for writing by ensuring that it is not deleted.

        Args:
            key (str): The key.
        """
        self._deletes.discard(key)

    def _is_pending_delete(self, key: str) -> bool:
        """
        Whether the given key is marked as deleted. Note, this is not the case
        after _prepare_key_for_write.

        Args:
            key (str): The key.

        Returns:
            bool: Whether the given key is in the delete set.
        """
        return key in self._deletes

    def _is_alive(self, value: tuple[str, float | None]) -> bool:
        """
        Checks whether the given value has expired.

        Args:
            value (tuple[str, float  |  None]): Tuple of value and optional
            expiration monotonic time point.

        Returns:
            bool: Whether the value has expired.
        """
        _, expire = value
        if expire is None:
            return True
        now_mono = time.monotonic()
        if expire > now_mono:
            return True
        return False

    def _clean_vals(self) -> None:
        """
        Cleans up expired values by deleting them.
        """
        if self._parent is not None:
            return
        remove = set()
        for key, value in self._vals.items():
            if self._is_alive(value):
                continue
            remove.add(key)
        self.delete(remove)

    def raw_vals(self) -> dict[str, tuple[str, float | None]]:
        """
        Raw access to the value dictionary.

        Returns:
            dict[str, tuple[str, float | None]]: The actual value dictionary of
            this state.
        """
        return self._vals

    def raw_queues(
            self) -> dict[str, collections.deque[str]]:
        """
        Raw access to the queues dictionary.

        Returns:
            dict[str, collections.deque[str]]: The actual queue dictionary of
            this state.
        """
        return self._queues

    def raw_hashes(self) -> dict[str, dict[str, str]]:
        """
        Raw access to the hashes dictionary.

        Returns:
            dict[str, dict[str, str]]: The actual hashes dictionary of this
            state.
        """
        return self._hashes

    def raw_zorder(self) -> dict[str, list[str]]:
        """
        Raw access to the zorder dictionary.

        Returns:
            dict[str, list[str]]: The actual zorder dictionary of this state.
        """
        return self._zorder

    def raw_zscores(self) -> dict[str, dict[str, float]]:
        """
        Raw access to the zscores dictionary.

        Returns:
            dict[str, dict[str, float]]: The actual zscores dictionary of this
            state.
        """
        return self._zscores

    def raw_deletes(self) -> set[str]:
        """
        Raw access to the deletes set.

        Returns:
            set[str]: The actual deletes set of this state.
        """
        return self._deletes

    def set_value(self, key: str, value: str, expire: float | None) -> None:
        """
        Sets the value for the given key.

        Args:
            key (str): The key.
            value (str): The value.
            expire (float | None): When the value should expire if at all.
        """
        if key not in self._vals:
            self.verify_key("value", key)
        self._prepare_key_for_write(key)
        self._vals[key] = (value, expire)

    def get_value(self, key: str) -> tuple[str, float | None] | None:
        """
        Retrieves the value associated with the given key.

        Args:
            key (str): The key.

        Returns:
            tuple[str, float | None] | None: The value of the key and when it
            will expire.
        """
        res = self._vals.get(key)
        if res is not None and not self._is_alive(res):
            self._clean_vals()
            return None
        if (
                res is None
                and self._parent is not None
                and not self._is_pending_delete(key)):
            return self._parent.get_value(key)
        return res

    def has_value(self, key: str) -> bool:
        """
        Whether the given key is associated with a value. Note, this is
        different from whether the key exists.

        Args:
            key (str): The key.

        Returns:
            bool: If the key exists and its type is "value".
        """
        return self.get_value(key) is not None

    def get_queue(self, key: str) -> collections.deque[str]:
        """
        Returns a queue for writing. If the queue doesn't exist already it will
        be created.

        Args:
            key (str): The key.

        Returns:
            collections.deque[str]: The queue for writing.
        """
        res = self._queues.get(key)
        if res is None:
            self.verify_key("list", key)
            if self._parent is not None and not self._is_pending_delete(key):
                res = collections.deque(self._parent.get_queue(key))
            else:
                res = collections.deque()
            self._queues[key] = res
        self._prepare_key_for_write(key)
        return res

    def readonly_queue(self, key: str) -> collections.deque[str] | None:
        """
        Returns a queue for reading only. If the queue doesn't exist None is
        returned.

        Args:
            key (str): The key.

        Returns:
            collections.deque[str] | None: The readonly queue. Make sure to not
            modify it as it is not a copy. None if the queue doesn't exist.
        """
        res = self._queues.get(key)
        if res is None:
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_queue(key)
        return res

    def queue_len(self, key: str) -> int:
        """
        Computes the length of a queue.

        Args:
            key (str): The key.

        Returns:
            int: The length of the queue.
        """
        res = self.readonly_queue(key)
        if res is None:
            return 0
        return len(res)

    def get_hash(self, key: str) -> dict[str, str]:
        """
        Returns a hash for writing. If the hash doesn't exist already it will
        be created.

        Args:
            key (str): The key.

        Returns:
            dict[str, str]: The hash for writing.
        """
        res = self._hashes.get(key)
        if res is None:
            self.verify_key("hash", key)
            if self._parent is not None and not self._is_pending_delete(key):
                res = dict(self._parent.get_hash(key))
            else:
                res = {}
            self._hashes[key] = res
        self._prepare_key_for_write(key)
        return res

    def readonly_hash(self, key: str) -> dict[str, str] | None:
        """
        Returns a hash for reading only. If the hash doesn't exist None is
        returned.

        Args:
            key (str): The key.

        Returns:
            dict[str, str] | None: The readonly hash. Make sure to not
            modify it as it is not a copy. None if the hash doesn't exist.
        """
        res = self._hashes.get(key)
        if res is None:
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_hash(key)
        return res

    def get_zset(self, key: str) -> tuple[list[str], dict[str, float]]:
        """
        Returns a zset for writing. If the zset doesn't exist already it
        will be created.

        Args:
            key (str): The key.

        Returns:
            tuple[list[str], dict[str, float]]: The tuple of zorder and zscores
            for writing.
        """
        rorder = self._zorder.get(key)
        rscores = self._zscores.get(key)
        if rorder is None:
            self.verify_key("zset", key)
            if self._parent is not None and not self._is_pending_delete(key):
                porder, pscores = self._parent.get_zset(key)
                rorder = list(porder)
                rscores = dict(pscores)
            else:
                rorder = []
                rscores = {}
            self._zorder[key] = rorder
            self._zscores[key] = rscores
        self._prepare_key_for_write(key)
        assert rscores is not None
        return (rorder, rscores)

    def readonly_zorder(self, key: str) -> list[str] | None:
        """
        Returns a zorder for reading only. If the zorder doesn't exist None is
        returned.

        Args:
            key (str): The key.

        Returns:
            list[str] | None: The readonly zorder. Make sure to not
            modify it as it is not a copy. None if the zorder doesn't exist.
        """
        res = self._zorder.get(key)
        if res is None:
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_zorder(key)
        return res

    def readonly_zscores(self, key: str) -> dict[str, float] | None:
        """
        Returns a zscores for reading only. If the zscores doesn't exist None
        is returned.

        Args:
            key (str): The key.

        Returns:
            dict[str, float] | None: The readonly zscores. Make sure to not
            modify it as it is not a copy. None if the zscores doesn't exist.
        """
        res = self._zscores.get(key)
        if res is None:
            if self._parent is None or self._is_pending_delete(key):
                return None
            return self._parent.readonly_zscores(key)
        return res

    def zorder_len(self, key: str) -> int:
        """
        Computes the length of a zorder.

        Args:
            key (str): The key.

        Returns:
            int: The length of the zorder.
        """
        res = self.readonly_zorder(key)
        if res is None:
            return 0
        return len(res)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}["
            f"vals={self._vals},"
            f"queues={dict(self._queues)},"
            f"hashes={dict(self._hashes)},"
            f"zorder={dict(self._zorder)},"
            f"zscores={dict(self._zscores)},"
            f"deletes={self._deletes},"
            f"has_parent={self._parent is not None}]")

    def __repr__(self) -> str:
        return self.__str__()


class Machine(RedisAPI):
    """
    A Machine manages the state of a memory runtime and exposes the redis API.
    """
    def __init__(self, state: State) -> None:
        """
        Creates a Machine.

        Args:
            state (State): The associated state.
        """
        super().__init__()
        self._state = state

    def get_state(self) -> State:
        """
        Returns the managed state.

        Returns:
            State: The associated state.
        """
        return self._state

    def exists(self, *keys: str) -> int:
        res = 0
        for key in keys:
            if self._state.exists(key):
                res += 1
        return res

    def delete(self, *keys: str) -> int:
        res = self.exists(*keys)
        self._state.delete(set(keys))
        return res

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[True],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> str | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[False],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> bool | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        ...

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        expire = compute_expire(
            expire_timestamp=expire_timestamp, expire_in=expire_in)
        prev_value = None
        prev_expire = None
        prev = self._state.get_value(key)
        if prev is not None:
            prev_value, prev_expire = prev
        if keep_ttl:
            expire = prev_expire
        do_set = False
        if mode == RSM_ALWAYS:
            do_set = True
        elif mode == RSM_EXISTS:
            do_set = prev is not None
        elif mode == RSM_MISSING:
            do_set = prev is None
        else:
            raise ValueError(f"unknown mode: {mode}")
        if do_set:
            self._state.set_value(key, value, expire)
        if return_previous:
            return prev_value
        return do_set

    def get(self, key: str) -> str | None:
        res = self._state.get_value(key)
        if res is None:
            return None
        value, _ = res
        return value

    def incrby(self, key: str, inc: float | int) -> float:
        res = self._state.get_value(key)
        if res is None:
            val = "0"
            expire = None
        else:
            val, expire = res
        num = float(val) + inc
        self._state.set_value(key, to_number_str(num), expire)
        return num

    def lpush(self, key: str, *values: str) -> int:
        queue = self._state.get_queue(key)
        queue.extendleft(values)
        return len(queue)

    def rpush(self, key: str, *values: str) -> int:
        queue = self._state.get_queue(key)
        queue.extend(values)
        return len(queue)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        queue = self._state.get_queue(key)
        if not queue:
            return None
        if count is None:
            rval = queue.popleft()
            if not queue:
                self.delete(key)
            return rval
        popc = count
        res = []
        while popc > 0 and queue:
            res.append(queue.popleft())
            popc -= 1
        if not queue:
            self.delete(key)
        return res if res else None

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        queue = self._state.get_queue(key)
        if not queue:
            return None
        if count is None:
            rval = queue.pop()
            if not queue:
                self.delete(key)
            return rval
        popc = count
        res = []
        while popc > 0 and queue:
            res.append(queue.pop())
            popc -= 1
        if not queue:
            self.delete(key)
        return res if res else None

    def lrange(self, key: str, start: int, stop: int) -> list[str]:
        queue = self._state.readonly_queue(key)
        if queue is None:
            return []
        if start >= len(queue):
            return []
        if start < 0:
            start = max(0, start + len(queue))
        if stop < 0:
            stop += len(queue)
            if stop < 0:
                return []
        stop += 1
        queue.rotate(-start)
        res = list(itertools.islice(queue, 0, stop - start, 1))
        queue.rotate(start)
        return res

    def llen(self, key: str) -> int:
        return self._state.queue_len(key)

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        count = 0
        zorder, zscores = self._state.get_zset(key)
        for name, score in mapping.items():
            if name not in zscores:
                zorder.append(name)
                count += 1
            zscores[name] = score
        zorder.sort(key=lambda k: (zscores[k], k))
        return count

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        zorder, zscores = self._state.get_zset(key)
        res = []
        remain = 1 if count is None else count
        while remain > 0 and zorder:
            name = zorder.pop()
            score = zscores.pop(name)
            res.append((name, score))
            remain -= 1
        return res

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        zorder, zscores = self._state.get_zset(key)
        res = []
        remain = 1 if count is None else count
        while remain > 0 and zorder:
            name = zorder.pop(0)
            score = zscores.pop(name)
            res.append((name, score))
            remain -= 1
        return res

    def zrange(self, key: str, start: int, stop: int) -> list[str]:
        zorder = self._state.readonly_zorder(key)
        if zorder is None:
            return []
        astop: int | None = stop
        if astop == -1 or astop is None:  # NOTE: mypy workaround
            astop = None
        else:
            astop += 1
        return zorder[start:astop]

    def zcard(self, key: str) -> int:
        return self._state.zorder_len(key)

    def hset(self, key: str, mapping: dict[str, str]) -> int:
        obj = self._state.get_hash(key)
        res = len(set(mapping.keys()).difference(obj.keys()))
        obj.update(mapping)
        return res

    def hdel(self, key: str, *fields: str) -> int:
        obj = self._state.get_hash(key)
        res = 0
        for field in fields:
            if obj.pop(field, None) is not None:
                res += 1
        if not obj:
            self._state.delete({key})
        return res

    def hget(self, key: str, field: str) -> str | None:
        res = self._state.readonly_hash(key)
        if res is None:
            return None
        return res.get(field)

    def hmget(self, key: str, *fields: str) -> dict[str, str | None]:
        res = self._state.readonly_hash(key)
        if res is None:
            return {}
        return {
            field: res.get(field)
            for field in fields
        }

    def hincrby(self, key: str, field: str, inc: float | int) -> float:
        res = self._state.get_hash(key)
        num = float(res.get(field, 0)) + inc
        res[field] = to_number_str(num)
        return num

    def hkeys(self, key: str) -> list[str]:
        res = self._state.readonly_hash(key)
        if res is None:
            return []
        return list(res.keys())

    def hvals(self, key: str) -> list[str]:
        res = self._state.readonly_hash(key)
        if res is None:
            return []
        return list(res.values())

    def hgetall(self, key: str) -> dict[str, str]:
        res = self._state.readonly_hash(key)
        if res is None:
            return {}
        return dict(res)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__name__}[state={self._state}]")

    def __repr__(self) -> str:
        return self.__str__()



"""Provides core functionality of expressions."""
from collections.abc import Callable

from redipy.graph.expr import ExprObj, ValueType


class Expr:  # pylint: disable=too-few-public-methods
    """The base class for all expressions."""
    def compile(self) -> ExprObj:
        """
        Compiles the expression into an execution graph expression object.

        Returns:
            ExprObj: The expression object.
        """
        raise NotImplementedError()

    def __add__(self, other: 'MixedType') -> 'Expr':
        """
        Adds two expressions.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return AddOp(self, other)

    def __sub__(self, other: 'MixedType') -> 'Expr':
        """
        Subtracts two expressions.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return SubOp(self, other)

    def eq_(self, other: 'MixedType') -> 'Expr':
        """
        Compares two expressions for equality.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return EqOp(self, other)

    def ne_(self, other: 'MixedType') -> 'Expr':
        """
        Compares two expressions for inequality.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return NeOp(self, other)

    def lt_(self, other: 'MixedType') -> 'Expr':
        """
        Compares two expressions returning whether this expression is less than
        the other expression.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return LtOp(self, other)

    def le_(self, other: 'MixedType') -> 'Expr':
        """
        Compares two expressions returning whether this expression is less or
        equal to the other expression.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return LeOp(self, other)

    def gt_(self, other: 'MixedType') -> 'Expr':
        """
        Compares two expressions returning whether this expression is greater
        than the other expression.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return GtOp(self, other)

    def ge_(self, other: 'MixedType') -> 'Expr':
        """
        Compares two expressions returning whether this expression is greater
        or equal to the other expression.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return GeOp(self, other)

    def not_(self) -> 'Expr':
        """
        Negates the current expression.

        Returns:
            Expr: The negated expression.
        """
        return NotOp(self)

    def or_(self, other: 'MixedType') -> 'Expr':
        """
        Logically ORs this expression with the other expression.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return OrOp(self, other)

    def and_(self, other: 'MixedType') -> 'Expr':
        """
        Logically ANDs this expression with the other expression.

        Args:
            other (MixedType): The other expression.

        Returns:
            Expr: The resulting expression.
        """
        return AndOp(self, other)


class ExprHelper(Expr):  # pylint: disable=too-few-public-methods
    """Provides a wrapper to use a function to get an expression."""
    def __init__(self, expr_fn: Callable[[], ExprObj]) -> None:
        """
        Creates a wrapper to use a function to get an expression.

        Args:
            expr_fn (Callable[[], ExprObj]): Function that returns an
            expression.
        """
        super().__init__()
        self._expr_fn = expr_fn

    def compile(self) -> ExprObj:
        return self._expr_fn()


class Constant(Expr):
    """Accesses a named constant."""
    def __init__(self, raw: str) -> None:
        """
        Accesses a named constant.

        Args:
            raw (str): The name of the constant.
        """
        super().__init__()
        self._raw = raw

    def compile(self) -> ExprObj:
        return {
            "kind": "constant",
            "raw": self._raw,
        }


LiteralType = str | int | float | bool | list | None
"""Literal values that transparently get converted to expressions."""
MixedType = LiteralType | Expr
"""An expression or literal."""


class Strs(Expr):
    """Concatenates a sequence of values."""
    def __init__(self, *values: MixedType) -> None:
        """
        Concatenates a sequence of values.

        Args:
            *values (MixedType): The values to concatenate. Should be strings.
        """
        super().__init__()
        self._values = [lit_helper(val) for val in values]

    def compile(self) -> ExprObj:
        return {
            "kind": "concat",
            "strings": [val.compile() for val in self._values],
        }


class LiteralOp(Expr):
    """Expression for literal values."""
    def __init__(self, value: LiteralType) -> None:
        """
        Expression for literal values.

        Args:
            value (LiteralType): The literal value.
        """
        super().__init__()
        self._value = value
        self._type = self.compute_type(value)

    @staticmethod
    def compute_type(value: LiteralType) -> ValueType:
        """
        Determines the type of the literal.

        Args:
            value (LiteralType): The literal.

        Raises:
            ValueError: If the type can not be determined.

        Returns:
            ValueType: The type of the literal.
        """
        if value is None:
            return "none"
        if isinstance(value, bool):
            return "bool"
        if isinstance(value, int):
            return "int"
        if isinstance(value, float):
            return "float"
        if isinstance(value, str):
            return "str"
        if isinstance(value, list):
            return "list"
        raise ValueError(f"unknown type for: {value}")

    def compile(self) -> ExprObj:
        return {
            "kind": "val",
            "value": self._value,
            "type": self._type,
        }


def lit_helper(value: MixedType) -> Expr:
    """
    Converts a mixed type into an expression. If the mixed type is a literal
    it will be wrapped in an expression.

    Args:
        value (MixedType): An expression or literal.

    Returns:
        Expr: The expression.
    """
    if isinstance(value, Expr):
        return value
    return LiteralOp(value)


class NotOp(Expr):
    """Negates an expression."""
    def __init__(self, expr: MixedType) -> None:
        """
        Negates an expression. Do not call directly. Use the expression method
        instead.

        Args:
            expr (MixedType): The expression to negate.
        """
        super().__init__()
        self._expr = lit_helper(expr)

    def compile(self) -> ExprObj:
        return {
            "kind": "unary",
            "op": "not",
            "arg": self._expr.compile(),
        }


class Op(Expr):
    """Combines two expressions using an operation."""
    def __init__(self, lhs: MixedType, rhs: MixedType) -> None:
        """
        Combines two expressions using an operation. Do not call directly.
        Use the corresponding expression method instead.

        Args:
            lhs (MixedType): The left expression.

            rhs (MixedType): The right expression.
        """
        super().__init__()
        self._lhs = lit_helper(lhs)
        self._rhs = lit_helper(rhs)

    def get_left(self) -> Expr:
        """
        Returns the left expression.

        Returns:
            Expr: The left expression.
        """
        return self._lhs

    def get_right(self) -> Expr:
        """
        Returns the right expression.

        Returns:
            Expr: The right expression.
        """
        return self._rhs

    def compile(self) -> ExprObj:
        raise NotImplementedError()


class AndOp(Op):
    """Computes the logical AND of both expressions."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "and",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class OrOp(Op):
    """Computes the logical OR of both expressions."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "or",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class AddOp(Op):
    """Adds two expressions."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "add",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class SubOp(Op):
    """Subtracts two expressions."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "sub",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class LtOp(Op):
    """Computes whether the left hand side is less than the right hand side."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "lt",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class LeOp(Op):
    """Computes whether the left hand side is less or equal to the right hand
    side."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "le",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class GtOp(Op):
    """Computes whether the left hand side is greater than the right hand
    side."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "gt",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class GeOp(Op):
    """Computes whether the left hand side is greater or equal to the right
    hand side."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "ge",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class EqOp(Op):
    """Computes whether two expressions evaluate to equal values."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "eq",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }


class NeOp(Op):
    """Computes whether two expressions evaluate to unequal values."""
    def compile(self) -> ExprObj:
        return {
            "kind": "binary",
            "op": "ne",
            "left": self.get_left().compile(),
            "right": self.get_right().compile(),
        }



"""This module defines the basic redis API. All redis functions appear once
in RedisAPI and once in PipelineAPI. Additional functionality is added via
RedisClientAPI."""
import contextlib
import datetime
from collections.abc import Iterator
from typing import Literal, overload

from redipy.backend.backend import ExecFunction
from redipy.symbolic.seq import FnContext


RSetMode = Literal[
    "always",
    "if_missing",  # NX
    "if_exists",  # XX
]
"""The conditions on when to set a value for the set command."""
RSM_ALWAYS: RSetMode = "always"
"""The value will always be set."""
RSM_MISSING: RSetMode = "if_missing"
"""The value will only be set when the key was missing.
This is equivalent to the NX flag."""
RSM_EXISTS: RSetMode = "if_exists"
"""The value will only be set when the key did exist.
This is equivalent to the XX flag."""


class PipelineAPI:
    """Redis API as pipeline. All methods return None and you have to call
    execute to retrieve the results of the pipeline commands."""
    def execute(self) -> list:
        """
        Executes the pipeline and returns the result values of each command.

        Returns:
            list: The result values of each command.
        """
        raise NotImplementedError()

    def exists(self, *keys: str) -> None:
        """
        Determines whether specified keys exist.

        See also the redis documentation: https://redis.io/commands/exists/

        The pipeline value is set to the number of keys that exist.

        Args:
            *keys (str): The keys.
        """
        raise NotImplementedError()

    def delete(self, *keys: str) -> None:
        """
        Deletes keys.

        See also the redis documentation: https://redis.io/commands/del/

        The pipeline value is set to the number of keys that got removed.

        Args:
            *keys (str): The keys.
        """
        raise NotImplementedError()

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> None:
        """
        Sets a value for a given key. The value can be scheduled to expire.

        See also the redis documentation: https://redis.io/commands/set/

        The pipeline value depends on the return_previous argument.

        Args:
            key (str): The key.

            value (str): The value.

            mode (RSetMode, optional): Under which condition to set the value
            valid values are RSM_ALWAYS, RSM_MISSING, and RSM_EXISTS.
            RSM_MISSING is the equivalent of setting the NX flag. RSM_EXISTS is
            the equivalent of the XX flag. Defaults to RSM_ALWAYS.

            return_previous (bool, optional): Whether to return the previous
            value associated with the key. Defaults to False.

            expire_timestamp (datetime.datetime | None, optional): A timestamp
            on when to expire the key. Defaults to None.

            expire_in (float | None, optional): A relative time in seconds on
            when to expire the key. Defaults to None.

            keep_ttl (bool, optional): Whether to keep previous expiration
            times. Defaults to False.
        """
        raise NotImplementedError()

    def get(self, key: str) -> None:
        """
        Retrieves the value for the given key.

        See also the redis documentation: https://redis.io/commands/get/

        The pipeline value is the value or None if the key does not exists or
        the value has expired.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def incrby(self, key: str, inc: float | int) -> None:
        """
        Updates the value associated with the given key by a relative amount.
        The value is interpreted as number. If the value doesn't exist zero is
        used as starting point.

        See also the redis documentation:
        https://redis.io/commands/incrby/
        https://redis.io/commands/incrbyfloat/

        The pipeline value is set to the new value as float.
        If the value cannot be interpreted as float while executing the
        pipeline a ValueError exception is raised.

        Args:
            key (str): The key.

            inc (float | int): The relative change.
        """
        raise NotImplementedError()

    def lpush(self, key: str, *values: str) -> None:
        """
        Pushes values to the left side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/lpush/

        The pipeline value is the length of the list after the push.

        Args:
            key (str): The key.

            *values (str): The values to push.
        """
        raise NotImplementedError()

    def rpush(self, key: str, *values: str) -> None:
        """
        Pushes values to the right side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/rpush/

        The pipeline value is the length of the list after the push.

        Args:
            key (str): The key.

            *values (str): The values to push.
        """
        raise NotImplementedError()

    def lpop(
            self,
            key: str,
            count: int | None = None) -> None:
        """
        Pops a number of values from the left side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/lpop/

        The pipeline value is None if the key doesn't exist. If a count
        is set a list with values in pop order is set as pipeline value (even
        if it is set to one). If count is not set (default or None) the single
        value that got popped is set as pipeline value.

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.
        """
        raise NotImplementedError()

    def rpop(
            self,
            key: str,
            count: int | None = None) -> None:
        """
        Pops a number of values from the right side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/rpop/

        The pipeline value is None if the key doesn't exist. If a count
        is set a list with values in pop order is set as pipeline value (even
        if it is set to one). If count is not set (default or None) the single
        value that got popped is set as pipeline value.

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.
        """
        raise NotImplementedError()

    def lrange(self, key: str, start: int, stop: int) -> None:
        """
        Returns a number of values from the list specified by the given range.
        Negative numbers are interpreted as index from the back of the list.
        Out of range indices are ignored, potentially returning an empty list.

        See also the redis documentation: https://redis.io/commands/lrange/

        The pipeline value is the resulting elements.

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).
        """
        raise NotImplementedError()

    def llen(self, key: str) -> None:
        """
        Computes the length of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/llen/

        The length of the list is set as pipeline value.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def zadd(self, key: str, mapping: dict[str, float]) -> None:
        """
        Adds elements to the sorted set associated with the key.

        See also the redis documentation: https://redis.io/commands/zadd/

        NOTE: not all setting modes are implemented yet.

        The number of new members is set as pipeline value.

        Args:
            key (str): The key.
            mapping (dict[str, float]): A dictionary with values and scores.
        """
        raise NotImplementedError()

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        """
        Pops a number of members of the sorted set associated with the given
        key with the highest scores.

        See also the redis documentation: https://redis.io/commands/zpopmax/

        The members with their associated scores in pop order is set as
        pipeline value.

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.
        """
        raise NotImplementedError()

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        """
        Pops a number of members of the sorted set associated with the given
        key with the lowest scores.

        See also the redis documentation: https://redis.io/commands/zpopmin/

        The members with their associated scores in pop order is set as
        pipeline value.

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.
        """
        raise NotImplementedError()

    def zrange(self, key: str, start: int, stop: int) -> None:
        """
        Returns a number of values from the sorted set specified by the given
        range. As of now the indices are based on the order of the set.
        Negative numbers are interpreted as index from the back of the set.
        Out of range indices are ignored, potentially returning an empty set.

        See also the redis documentation: https://redis.io/commands/zrange/

        NOTE: not all modes are implemented yet.

        The members names are set as pipeline value.

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).
        """
        raise NotImplementedError()

    def zcard(self, key: str) -> None:
        """
        Computes the cardinality of the sorted set associated with the given
        key.

        See also the redis documentation: https://redis.io/commands/zcard/

        The number of members in the set is set as pipeline value.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def hset(self, key: str, mapping: dict[str, str]) -> None:
        """
        Sets a mapping for the given hash.

        See also the redis documentation: https://redis.io/commands/hset/

        The pipeline value is set to the number of fields added.

        Args:
            key (str): The key.

            mapping (dict[str, str]): The field value pairs to be set.
        """
        raise NotImplementedError()

    def hdel(self, key: str, *fields: str) -> None:
        """
        Deletes fields from the given hash.

        See also the redis documentation: https://redis.io/commands/hdel/

        The pipeline value is set to the number of fields that got deleted
        (excluding fields that did not exist).

        Args:
            key (str): The key.

            *fields (str): The fields to delete.
        """
        raise NotImplementedError()

    def hget(self, key: str, field: str) -> None:
        """
        Retrieves the value associated with a field of a hash.

        See also the redis documentation: https://redis.io/commands/hget/

        The pipeline value is set to the value of the field or None if the
        field doesn't exist.

        Args:
            key (str): The key.

            field (str): The field.
        """
        raise NotImplementedError()

    def hmget(self, key: str, *fields: str) -> None:
        """
        Retrieves the values associated with given fields of a hash.

        See also the redis documentation: https://redis.io/commands/hmget/

        The pipeline value is set to a dictionary with fields mapping to their
        values. If a field doesn't exist in the hash the value is returned
        as None.

        Args:
            key (str): The key.

            *fields (str): The fields to retrieve.
        """
        raise NotImplementedError()

    def hincrby(self, key: str, field: str, inc: float | int) -> None:
        """
        Interprets a field value of a hash as number and updates the value.

        See also the redis documentation:
        https://redis.io/commands/hincrby/
        https://redis.io/commands/hincrbyfloat/

        The pipeline value is set to the new value of the field.

        Args:
            key (str): The key.

            field (str): The field to interpret as number.

            inc (float | int): The relative numerical change.
        """
        raise NotImplementedError()

    def hkeys(self, key: str) -> None:
        """
        Retrieves the fields of a hash.

        See also the redis documentation: https://redis.io/commands/hkeys/

        The pipeline value is set to a list of all fields of the given hash.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def hvals(self, key: str) -> None:
        """
        Retrieves the values of a hash.

        See also the redis documentation: https://redis.io/commands/hvals/

        The pipeline value is set to a list of all values of the given hash.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()

    def hgetall(self, key: str) -> None:
        """
        Retrieves all fields and values of a hash.

        See also the redis documentation: https://redis.io/commands/hgetall/

        The pipeline value is set to a dictionary with fields mapping to their
        values.

        Args:
            key (str): The key.
        """
        raise NotImplementedError()


class RedisAPI:
    """The redis API."""
    def exists(self, *keys: str) -> int:
        """
        Determines whether specified keys exist.

        See also the redis documentation: https://redis.io/commands/exists/

        Args:
            *keys (str): The keys.

        Returns:
            int: The number of keys that exist.
        """
        raise NotImplementedError()

    def delete(self, *keys: str) -> int:
        """
        Deletes keys.

        See also the redis documentation: https://redis.io/commands/del/

        Args:
            *keys (str): The keys.

        Returns:
            int: The number of keys that got removed.
        """
        raise NotImplementedError()

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[True],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> str | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[False],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> bool | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        ...

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        """
        Sets a value for a given key. The value can be scheduled to expire.

        See also the redis documentation: https://redis.io/commands/set/

        Args:
            key (str): The key.

            value (str): The value.

            mode (RSetMode, optional): Under which condition to set the value
            valid values are RSM_ALWAYS, RSM_MISSING, and RSM_EXISTS.
            RSM_MISSING is the equivalent of setting the NX flag. RSM_EXISTS is
            the equivalent of the XX flag. Defaults to RSM_ALWAYS.

            return_previous (bool, optional): Whether to return the previous
            value associated with the key. Defaults to False.

            expire_timestamp (datetime.datetime | None, optional): A timestamp
            on when to expire the key. Defaults to None.

            expire_in (float | None, optional): A relative time in seconds on
            when to expire the key. Defaults to None.

            keep_ttl (bool, optional): Whether to keep previous expiration
            times. Defaults to False.

        Returns:
            str | bool | None: The return value depends on the return_previous
            argument.
        """
        raise NotImplementedError()

    def get(self, key: str) -> str | None:
        """
        Retrieves the value for the given key.

        See also the redis documentation: https://redis.io/commands/get/

        Args:
            key (str): The key.

        Returns:
            str | None: The value or None if the key does not exists or the
            value has expired.
        """
        raise NotImplementedError()

    def incrby(self, key: str, inc: float | int) -> float:
        """
        Updates the value associated with the given key by a relative amount.
        The value is interpreted as number. If the value doesn't exist zero is
        used as starting point.

        See also the redis documentation:
        https://redis.io/commands/incrby/
        https://redis.io/commands/incrbyfloat/

        Args:
            key (str): The key.

            inc (float | int): The relative change.

        Raises:
            ValueError: If the value cannot be interpreted as float.

        Returns:
            float: The new value as float.
        """
        raise NotImplementedError()

    def lpush(self, key: str, *values: str) -> int:
        """
        Pushes values to the left side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/lpush/

        Args:
            key (str): The key.

            *values (str): The values to push.

        Returns:
            int: The length of the list after the push.
        """
        raise NotImplementedError()

    def rpush(self, key: str, *values: str) -> int:
        """
        Pushes values to the right side of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/rpush/

        Args:
            key (str): The key.

            *values (str): The values to push.

        Returns:
            int: The length of the list after the push.
        """
        raise NotImplementedError()

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        """
        Pops a number of values from the left side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/lpop/

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.

        Returns:
            str | list[str] | None: None if the key doesn't exist. If a count
            is set a list with values in pop order is returned (even if it is
            set to one). If count is not set (default or None) the single value
            that got popped is returned.
        """
        raise NotImplementedError()

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        """
        Pops a number of values from the right side of the list associated with
        the key.

        See also the redis documentation: https://redis.io/commands/rpop/

        Args:
            key (str): The key.

            count (int | None, optional): The number values to pop.
            Defaults to a single value.

        Returns:
            str | list[str] | None: None if the key doesn't exist. If a count
            is set a list with values in pop order is returned (even if it is
            set to one). If count is not set (default or None) the single value
            that got popped is returned.
        """
        raise NotImplementedError()

    def lrange(self, key: str, start: int, stop: int) -> list[str]:
        """
        Returns a number of values from the list specified by the given range.
        Negative numbers are interpreted as index from the back of the list.
        Out of range indices are ignored, potentially returning an empty list.

        See also the redis documentation: https://redis.io/commands/lrange/

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).

        Returns:
            list[str]: The elements.
        """
        raise NotImplementedError()

    def llen(self, key: str) -> int:
        """
        Computes the length of the list associated with the key.

        See also the redis documentation: https://redis.io/commands/llen/

        Args:
            key (str): The key.

        Returns:
            int: The length of the list.
        """
        raise NotImplementedError()

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        """
        Adds elements to the sorted set associated with the key.

        See also the redis documentation: https://redis.io/commands/zadd/

        NOTE: not all setting modes are implemented yet.

        Args:
            key (str): The key.
            mapping (dict[str, float]): A dictionary with values and scores.

        Returns:
            int: The number of new members.
        """
        raise NotImplementedError()

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        """
        Pops a number of members of the sorted set associated with the given
        key with the highest scores.

        See also the redis documentation: https://redis.io/commands/zpopmax/

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.

        Returns:
            list[tuple[str, float]]: The members with their associated scores
            in pop order.
        """
        raise NotImplementedError()

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        """
        Pops a number of members of the sorted set associated with the given
        key with the lowest scores.

        See also the redis documentation: https://redis.io/commands/zpopmin/

        Args:
            key (str): The key.

            count (int, optional): The number of members to remove.
            Defaults to 1.

        Returns:
            list[tuple[str, float]]: The members with their associated scores
            in pop order.
        """
        raise NotImplementedError()

    def zrange(self, key: str, start: int, stop: int) -> list[str]:
        """
        Returns a number of values from the sorted set specified by the given
        range. As of now the indices are based on the order of the set.
        Negative numbers are interpreted as index from the back of the set.
        Out of range indices are ignored, potentially returning an empty set.

        See also the redis documentation: https://redis.io/commands/zrange/

        NOTE: not all modes are implemented yet.

        Args:
            key (str): The key.

            start (int): The start index.

            stop (int): The stop index (inclusive).

        Returns:
            list[str]: The members names.
        """
        raise NotImplementedError()

    def zcard(self, key: str) -> int:
        """
        Computes the cardinality of the sorted set associated with the given
        key.

        See also the redis documentation: https://redis.io/commands/zcard/

        Args:
            key (str): The key.

        Returns:
            int: The number of members in the set.
        """
        raise NotImplementedError()

    def hset(self, key: str, mapping: dict[str, str]) -> int:
        """
        Sets a mapping for the given hash.

        See also the redis documentation: https://redis.io/commands/hset/

        Args:
            key (str): The key.

            mapping (dict[str, str]): The field value pairs to be set.

        Returns:
            int: The number of fields added.
        """
        raise NotImplementedError()

    def hdel(self, key: str, *fields: str) -> int:
        """
        Deletes fields from the given hash.

        See also the redis documentation: https://redis.io/commands/hdel/

        Args:
            key (str): The key.

            *fields (str): The fields to delete.

        Returns:
            int: The number of fields that got deleted (excluding fields that
            did not exist).
        """
        raise NotImplementedError()

    def hget(self, key: str, field: str) -> str | None:
        """
        Retrieves the value associated with a field of a hash.

        See also the redis documentation: https://redis.io/commands/hget/

        Args:
            key (str): The key.

            field (str): The field.

        Returns:
            str | None: The value of the field or None if the field doesn't
            exist.
        """
        raise NotImplementedError()

    def hmget(self, key: str, *fields: str) -> dict[str, str | None]:
        """
        Retrieves the values associated with given fields of a hash.

        See also the redis documentation: https://redis.io/commands/hmget/

        Args:
            key (str): The key.

            *fields (str): The fields to retrieve.

        Returns:
            dict[str, str | None]: A dictionary with fields mapping to their
            values. If a field doesn't exist in the hash the value is returned
            as None.
        """
        raise NotImplementedError()

    def hincrby(self, key: str, field: str, inc: float | int) -> float:
        """
        Interprets a field value of a hash as number and updates the value.

        See also the redis documentation:
        https://redis.io/commands/hincrby/
        https://redis.io/commands/hincrbyfloat/

        Args:
            key (str): The key.

            field (str): The field to interpret as number.

            inc (float | int): The relative numerical change.

        Returns:
            float: The new value of the field.
        """
        raise NotImplementedError()

    def hkeys(self, key: str) -> list[str]:
        """
        Retrieves the fields of a hash.

        See also the redis documentation: https://redis.io/commands/hkeys/

        Args:
            key (str): The key.

        Returns:
            list[str]: All fields of the given hash.
        """
        raise NotImplementedError()

    def hvals(self, key: str) -> list[str]:
        """
        Retrieves the values of a hash.

        See also the redis documentation: https://redis.io/commands/hvals/

        Args:
            key (str): The key.

        Returns:
            list[str]: All values of the given hash.
        """
        raise NotImplementedError()

    def hgetall(self, key: str) -> dict[str, str]:
        """
        Retrieves all fields and values of a hash.

        See also the redis documentation: https://redis.io/commands/hgetall/

        Args:
            key (str): The key.

        Returns:
            dict[str, str]: A dictionary with fields mapping to their values.
        """
        raise NotImplementedError()


class RedisClientAPI(RedisAPI):
    """This class enriches the redis API with pipeline and script
    functionality."""
    @contextlib.contextmanager
    def pipeline(self) -> Iterator[PipelineAPI]:
        """
        Starts a redis pipeline. When leaving the resource block the pipeline
        is executed automatically and the results are discarded. If you need
        the results call execute on the pipeline object.

        Yields:
            PipelineAPI: The pipeline.
        """
        raise NotImplementedError()

    def register_script(self, ctx: FnContext) -> ExecFunction:
        """
        Registers a script that can be executed in this redis runtime.

        Args:
            ctx (FnContext): The script to register.

        Returns:
            ExecFunction: A python that can be called to execute the script.
        """
        raise NotImplementedError()



"""This module provides core functionality for the symbolic scripting
engine."""
from collections.abc import Callable

from redipy.graph.cmd import AssignmentObj, CommandObj
from redipy.graph.expr import ArgObj, ExprObj, KeyObj, RefIdObj, VarObj
from redipy.symbolic.expr import Expr, ExprHelper, lit_helper, MixedType


class Compilable:  # pylint: disable=too-few-public-methods
    """Can be compiled to a CommandObj statement."""
    def compile(self) -> CommandObj:
        """
        Compiles to a statement.

        Returns:
            CommandObj: The CommandObj statement.
        """
        raise NotImplementedError()


class CmdHelper(Compilable):  # pylint: disable=too-few-public-methods
    """A wrapper to allow compiling anonymously."""
    def __init__(self, stmt_fn: Callable[[], CommandObj]) -> None:
        """
        Creates a command helper.

        Args:
            stmt_fn (Callable[[], CommandObj]): The compiling function.
        """
        self._stmt_fn = stmt_fn

    def compile(self) -> CommandObj:
        return self._stmt_fn()


class Variable(Expr):
    """A reference to a variable. Various methods to interact with the variable
    are available."""
    def __init__(self) -> None:
        self._index: int | None = None

    def len_(self) -> 'Expr':
        """
        Computes the length of the array assuming that the variable points to
        one.

        Returns:
            Expr: The expression to compute the length.
        """
        return ArrayLen(self)

    def __getitem__(self, other: 'MixedType') -> 'Expr':
        """
        Accesses an element of the array assuming that the variable points to
        one.

        Args:
            other (MixedType): The index.

        Returns:
            Expr: The expression to access an index.
        """
        return ArrayAt(self, other)

    def set_index(self, index: int) -> None:
        """
        Sets the internal number of this variable.

        Args:
            index (int): The index of this variable.
        """
        self._index = index

    def get_index(self) -> int:
        """
        Returns the internal number of this variable.

        Returns:
            int: The index of this variable.
        """
        assert self._index is not None
        return self._index

    def prefix(self) -> str:
        """
        The name prefix of this variable.

        Returns:
            str: The name prefix.
        """
        raise NotImplementedError()

    def get_declare(self) -> AssignmentObj:
        """
        Creates a statement that declares the variable. Do not call this method
        directly.

        Returns:
            AssignmentObj: The assignment statement.
        """
        return {
            "kind": "declare",
            "assign": self.get_ref(),
            "value": self.get_declare_rhs().compile(),
        }

    def get_declare_rhs(self) -> Expr:
        """
        Returns the expression containing the initial value of the variable.

        Returns:
            Expr: The initial value expression.
        """
        raise NotImplementedError()

    def get_ref(self) -> RefIdObj:
        """
        Get an expression to reference the value of the variable.

        Returns:
            RefIdObj: The reference expression.
        """
        raise NotImplementedError()

    def compile(self) -> RefIdObj:
        return self.get_ref()

    def assign(self, val: MixedType) -> CmdHelper:
        """
        Assigns a value to the variable.

        Args:
            val (MixedType): The value expression.

        Returns:
            CmdHelper: The statement to assign the value.
        """
        expr = lit_helper(val)
        return CmdHelper(
            lambda: {
                "kind": "assign",
                "assign": self.get_ref(),
                "value": expr.compile(),
            })

    def set_at(self, index: MixedType, val: MixedType) -> CmdHelper:
        """
        Sets the value at a given index of the array assuming the variable
        points to one.

        Args:
            index (MixedType): The index.
            val (MixedType): The value to assign.

        Returns:
            CmdHelper: The assignment statement.
        """
        ix = lit_helper(index)
        expr = lit_helper(val)
        return CmdHelper(
            lambda: {
                "kind": "assign_at",
                "assign": self.get_ref(),
                "index": ix.compile(),
                "value": expr.compile(),
            })


class KeyVariable(Variable):
    """A key argument variable."""
    def __init__(self, name: str) -> None:
        """
        A key argument variable. Do not call this directly.

        Args:
            name (str): The name of the key argument.
        """
        super().__init__()
        self._name = name

    def get_declare_rhs(self) -> Expr:
        return ExprHelper(
            lambda: {
                "kind": "load_key_arg",
                "index": self.get_index(),
            })

    def prefix(self) -> str:
        return "key"

    def get_key_ref(self) -> KeyObj:
        """
        Returns a reference to the key.

        Returns:
            KeyObj: The key reference.
        """
        return {
            "kind": "key",
            "name": f"{self.prefix()}_{self.get_index()}",
            "readable": self._name,
        }

    def get_ref(self) -> RefIdObj:
        return self.get_key_ref()


class JSONArg(Variable):
    """An argument to the script. This can contain any value that can be
    converted to JSON. Any value is converted to JSON for transport to the
    function and automatically converted back inside the script."""
    def __init__(self, name: str) -> None:
        """
        Creates a JSON argument. Do not call this directly.

        Args:
            name (str): The name of the argument.
        """
        super().__init__()
        self._name = name

    def get_declare_rhs(self) -> Expr:
        return ExprHelper(
            lambda: {
                "kind": "load_json_arg",
                "index": self.get_index(),
            })

    def prefix(self) -> str:
        return "arg"

    def get_arg_ref(self) -> ArgObj:
        """
        Get a reference to the argument.

        Returns:
            ArgObj: The reference expression.
        """
        return {
            "kind": "arg",
            "name": f"{self.prefix()}_{self.get_index()}",
            "readable": self._name,
        }

    def get_ref(self) -> RefIdObj:
        return self.get_arg_ref()


class LocalVariable(Variable):
    """A local variable."""
    def __init__(self, init: MixedType) -> None:
        """
        Creates a local variable. Do not call this directly.

        Args:
            init (MixedType): The initial value of the variable.
        """
        super().__init__()
        self._init = lit_helper(init)

    def get_declare_rhs(self) -> Expr:
        return self._init

    def prefix(self) -> str:
        return "var"

    def get_var_ref(self) -> VarObj:
        """
        Get a reference to the variable.

        Returns:
            VarObj: The reference expression.
        """
        return {
            "kind": "var",
            "name": f"{self.prefix()}_{self.get_index()}",
        }

    def get_ref(self) -> RefIdObj:
        return self.get_var_ref()


class ArrayAt(Expr):
    """Accesses an array at a given index."""
    def __init__(self, array: Variable, index: MixedType) -> None:
        """
        Accesses an array at the given index.

        Args:
            array (Variable): The array.
            index (MixedType): The index.
        """
        super().__init__()
        self._array = array
        self._index = lit_helper(index)

    def compile(self) -> ExprObj:
        return {
            "kind": "array_at",
            "var": self._array.get_ref(),
            "index": self._index.compile(),
        }


class ArrayLen(Expr):
    """Computes the length of an array."""
    def __init__(self, array: Variable) -> None:
        """
        Computes the length of an array.

        Args:
            array (Variable): The array.
        """
        super().__init__()
        self._array = array

    def compile(self) -> ExprObj:
        return {
            "kind": "array_len",
            "var": self._array.get_ref(),
        }



"""Defines all execution graph node types of statements. A statement can be
executed alone and usually has side-effects."""
from typing import Literal, TYPE_CHECKING, TypedDict

from redipy.graph.expr import ExprObj, IndexObj, RefIdObj, VarObj


if TYPE_CHECKING:
    from redipy.graph.seq import SequenceObj


AssignmentObj = TypedDict('AssignmentObj', {
    "kind": Literal["assign", "declare"],
    "assign": RefIdObj,
    "value": ExprObj,
})
"""Assigns an expression to a reference."""
AssignAtObj = TypedDict('AssignAtObj', {
    "kind": Literal["assign_at"],
    "assign": RefIdObj,
    "index": ExprObj,
    "value": ExprObj,
})
"""Assigns an expression to an index of a reference."""
StmtObj = TypedDict('StmtObj', {
    "kind": Literal["stmt"],
    "expr": ExprObj,
})
"""Executes an expression as statement."""
BranchObj = TypedDict('BranchObj', {
    "kind": Literal["branch"],
    "condition": ExprObj,
    "then": 'SequenceObj',
    "else": 'SequenceObj',
})
"""A branch with subsequences whether an expression was truthy or falsey."""
ForLoopObj = TypedDict('ForLoopObj', {
    "kind": Literal["for"],
    "array": ExprObj,
    "index": IndexObj,
    "value": VarObj,
    "body": 'SequenceObj',
})
"""A loop iterating through an array."""
WhileObj = TypedDict('WhileObj', {
    "kind": Literal["while"],
    "condition": ExprObj,
    "body": 'SequenceObj',
})
"""A loop executing until the condition fails."""
ReturnObj = TypedDict('ReturnObj', {
    "kind": Literal["return"],
    "value": ExprObj | None,
})
"""Sets the return value of the current stack frame."""
CommandObj = (
    AssignmentObj
    | AssignAtObj
    | StmtObj
    | BranchObj
    | ForLoopObj
    | WhileObj
    | ReturnObj
)
"""Defines all statements."""



"""Lua helper functions for patching lua function calls."""
from redipy.plugin import HelperFunction


class HPairlistScoresFn(HelperFunction):
    """Converts an alternating list of keys and values into a list of pairs."""
    @staticmethod
    def name() -> str:
        return "pairlist_scores"

    @staticmethod
    def args() -> str:
        return "arr"

    @staticmethod
    def body() -> str:
        return r"""
            local res = {}
            local key = nil
            for ix, value in ipairs(arr) do
                if ix % 2 == 1 then
                    key = value
                else
                    res[#res + 1] = {key, tonumber(value)}
                end
            end
            return res
        """


class HPairlistDictFn(HelperFunction):
    """Converts an alternating list of keys and values into a dictionary."""
    @staticmethod
    def name() -> str:
        return "pairlist_dict"

    @staticmethod
    def args() -> str:
        return "arr"

    @staticmethod
    def body() -> str:
        return r"""
            local res = {}
            local key = nil
            for _, value in ipairs(arr) do
                if key ~= nil then
                    res[key] = value
                    key = nil
                else
                    key = value
                end
            end
            return res
        """


class HKeyValDictFn(HelperFunction):
    """Constructs a dictionary from values and variadic keys."""
    @staticmethod
    def name() -> str:
        return "keyval_dict"

    @staticmethod
    def args() -> str:
        return "values, ..."

    @staticmethod
    def body() -> str:
        return r"""
            local res = {}
            for ix, key in ipairs(arg) do
                res[key] = values[ix] or cjson.null
            end
            return res
        """


class HNilOrIndexFn(HelperFunction):
    """Updates an index from 1-based (lua) to 0-based (python) if the value is
    not None (nil)."""
    @staticmethod
    def name() -> str:
        return "nil_or_index"

    @staticmethod
    def args() -> str:
        return "val"

    @staticmethod
    def body() -> str:
        return r"""
            if val ~= nil then
                val = val - 1
            end
            return val
        """


class HAsIntStrFn(HelperFunction):
    """Returns the integer value of a number."""
    @staticmethod
    def name() -> str:
        return "asintstr"

    @staticmethod
    def args() -> str:
        return "val"

    @staticmethod
    def body() -> str:
        return r"""
            return math.floor(val)
        """



"""Defines all execution graph node types of sequences. A sequence groups
together statements."""
from typing import Literal, TypedDict

from redipy.graph.cmd import CommandObj


SeqObj = TypedDict('SeqObj', {
    "kind": Literal["seq"],
    "cmds": list[CommandObj],
})
"""A normal sequence that does not add a new stack frame."""
ScriptObj = TypedDict('ScriptObj', {
    "kind": Literal["script"],
    "cmds": list[CommandObj],
    "argv": list[str],
    "keyv": list[str],
})
"""The base sequence of the script."""
FunctionObj = TypedDict('FunctionObj', {
    "kind": Literal["function"],
    "cmds": list[CommandObj],
    "argc": int,
})
"""A sequence for defining custom local functions."""


SequenceObj = SeqObj | ScriptObj | FunctionObj
"""A sequence grouping together statements."""



"""Defines the intermediate representation of redipy scripts. The execution
graph has three components: 1) expressions which output values, 2) statements
which consist mostly of expressions and typically change the execution context
via side-effects, and 3) sequences which group statements."""



"""The runtime for the redis backend."""
import contextlib
import datetime
import threading
import uuid
from collections.abc import Callable, Iterable, Iterator
from typing import (
    Any,
    cast,
    Literal,
    NoReturn,
    NotRequired,
    overload,
    Protocol,
    TypedDict,
)

from redis import Redis
from redis.client import Pipeline
from redis.commands.core import Script
from redis.exceptions import ResponseError

from redipy.api import (
    PipelineAPI,
    RSetMode,
    RSM_ALWAYS,
    RSM_EXISTS,
    RSM_MISSING,
)
from redipy.backend.runtime import Runtime
from redipy.redis.lua import LuaBackend
from redipy.util import (
    is_test,
    normalize_values,
    now,
    time_diff,
    to_list_str,
    to_maybe_str,
)


RedisConfig = TypedDict('RedisConfig', {
    "host": str,
    "port": int,
    "passwd": str,
    "prefix": NotRequired[str],
    "path": NotRequired[str],
})


class RedisFactory(Protocol):  # pylint: disable=too-few-public-methods
    """
    Factory function for creating a redis connection from a redis
    configuration.
    """
    def __call__(self, *, cfg: RedisConfig) -> Redis:
        """
        Creates a redis connection from a redis configuration.

        Args:
            cfg (RedisConfig): The redis configuration.

        Returns:
            Redis: The redis connection.
        """
        raise NotImplementedError()


class RedisFunctionBytes(Protocol):  # pylint: disable=too-few-public-methods
    """A redis function as it is returned from the redis connection."""
    def __call__(
            self,
            *,
            keys: list[str],
            args: list[Any],
            client: Redis | None) -> bytes:
        """
        Executes the redis function.

        Args:
            keys (list[str]): A list of keys used in the
            script. It is common convention to pass keys used in the script via
            this parameter but it is not necessary to do so. Importantly, keys
            can also be generated inside the script.

            args (list[Any]): Arguments to the script.

            client (Redis | None): An optionally different
            redis connection environment. Defaults to None.

        Returns:
            bytes: The result of the script.
        """
        raise NotImplementedError()


CONCURRENT_MODULE_CONN: int = 17
"""The number of allowed concurrent redis connections across threads."""


class RedisWrapper:
    """Manages redis connections."""
    def __init__(
            self,
            *,
            cfg: RedisConfig,
            redis_factory: RedisFactory | None = None,
            is_caching_enabled: bool = True) -> None:
        """
        Creates a redis wrapper for managing redis connections.

        Args:
            cfg (RedisConfig): The redis connection configuration.

            redis_factory (RedisFactory | None, optional): The redis connection
            factory. Defaults to None.

            is_caching_enabled (bool, optional): Whether caching of connections
            is enabled. Defaults to True.
        """
        self._redis_factory: RedisFactory = (
            self._create_connection
            if redis_factory is None else redis_factory)
        self._cfg = cfg
        self._is_caching_enabled = is_caching_enabled
        self._service_conn: list[Redis | None] = \
            [None] * CONCURRENT_MODULE_CONN
        self._lock = threading.RLock()

    @staticmethod
    def get_connection_index() -> int:
        """
        The connection index for this thread.

        Returns:
            int: The connection index.
        """
        return threading.get_ident() % CONCURRENT_MODULE_CONN

    @staticmethod
    def _create_connection(*, cfg: RedisConfig) -> Redis:
        return Redis(
            host=cfg["host"],
            port=cfg["port"],
            db=0,
            password=cfg["passwd"],
            retry_on_timeout=True,
            health_check_interval=45,
            client_name=f"rc-{uuid.uuid4().hex}")

    def _get_redis_cached_conn(self) -> Redis:
        if not self._is_caching_enabled:
            return self._redis_factory(cfg=self._cfg)

        ix = self.get_connection_index()
        res = self._service_conn[ix]
        if res is None:
            with self._lock:
                res = self._service_conn[ix]
                if res is None:
                    res = self._redis_factory(cfg=self._cfg)
                    self._service_conn[ix] = res
        return res

    @contextlib.contextmanager
    def get_connection(self) -> Iterator[Redis]:
        """
        Returns a redis connection.

        Yields:
            Redis: The redis connection.
        """
        success = False
        try:
            yield self._get_redis_cached_conn()
            success = True
        finally:
            if not success:
                self.reset()

    def reset(self) -> None:
        """
        Removes all cached connections.
        """
        with self._lock:
            for ix, _ in enumerate(self._service_conn):
                self._service_conn[ix] = None


class PipelineConnection(PipelineAPI):
    """A pipeline for the redis backend runtime."""
    def __init__(self, pipe: Pipeline, prefix: str) -> None:
        super().__init__()
        self._pipe = pipe
        self._fixes: list[Callable[[Any], Any]] = []
        self._prefix = prefix

    def get_pipeline(self) -> Pipeline:
        """
        Return the underlying redis pipeline.

        Returns:
            Pipeline: The redis pipeline.
        """
        return self._pipe

    def has_pending(self) -> bool:
        """
        Whether there are commands in the pipeline that have not been executed
        yet.

        Returns:
            bool: True if there are unexecuted commands in the pipeline.
        """
        return len(self._fixes) > 0

    def with_prefix(self, key: str) -> str:
        """
        Computes the actual key value of this redis instance.

        Args:
            key (str): The key.

        Returns:
            str: The actual key.
        """
        return f"{self._prefix}{key}"

    def add_fixup(self, fix: Callable[[Any], Any]) -> None:
        """
        Adds a fixup callback for the current pipeline command. Every pipeline
        command must have exactly one fixup callback (even if it is just an
        identity function).

        Args:
            fix (Callable[[Any], Any]): A callback to convert the pipeline
            command result into the correct output value.
        """
        self._fixes.append(fix)

    def execute(self) -> list:
        fixes = self._fixes
        self._fixes = []
        res = self._pipe.execute()
        assert len(res) == len(fixes)
        return [
            fixup(val)
            for val, fixup in zip(res, fixes)
        ]

    def exists(self, *keys: str) -> None:
        self._pipe.exists(*(
            self.with_prefix(key) for key in keys))
        self.add_fixup(int)

    def delete(self, *keys: str) -> None:
        self._pipe.delete(*(
            self.with_prefix(key) for key in keys))
        self.add_fixup(int)

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> None:
        expire = None
        if expire_in is not None:
            expire = int(expire_in * 1000.0)
        elif expire_timestamp is not None:
            expire = int(time_diff(now(), expire_timestamp) * 1000.0)
        self._pipe.set(
            self.with_prefix(key),
            value,
            get=return_previous,
            nx=(mode == RSM_MISSING),
            xx=(mode == RSM_EXISTS),
            px=expire,
            keepttl=keep_ttl)
        if return_previous:
            self.add_fixup(to_maybe_str)
        else:
            self.add_fixup(bool)

    def get(self, key: str) -> None:
        self._pipe.get(self.with_prefix(key))
        self.add_fixup(to_maybe_str)

    def incrby(self, key: str, inc: float | int) -> None:
        self._pipe.incrbyfloat(self.with_prefix(key), inc)
        self.add_fixup(float)

    def lpush(self, key: str, *values: str) -> None:
        self._pipe.lpush(self.with_prefix(key), *values)
        self.add_fixup(int)

    def rpush(self, key: str, *values: str) -> None:
        self._pipe.rpush(self.with_prefix(key), *values)
        self.add_fixup(int)

    def lpop(
            self,
            key: str,
            count: int | None = None) -> None:
        self._pipe.lpop(self.with_prefix(key), count)
        if count is None:
            self.add_fixup(to_maybe_str)
        else:
            self.add_fixup(to_list_str)

    def rpop(
            self,
            key: str,
            count: int | None = None) -> None:
        self._pipe.rpop(self.with_prefix(key), count)
        if count is None:
            self.add_fixup(to_maybe_str)
        else:
            self.add_fixup(to_list_str)

    def lrange(self, key: str, start: int, stop: int) -> None:
        self._pipe.lrange(self.with_prefix(key), start, stop)
        self.add_fixup(to_list_str)

    def llen(self, key: str) -> None:
        self._pipe.llen(self.with_prefix(key))
        self.add_fixup(int)

    def zadd(self, key: str, mapping: dict[str, float]) -> None:
        self._pipe.zadd(self.with_prefix(key), mapping=mapping)  # type: ignore
        self.add_fixup(int)

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        self._pipe.zpopmax(self.with_prefix(key), count)
        self.add_fixup(normalize_values)

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> None:
        self._pipe.zpopmin(self.with_prefix(key), count)
        self.add_fixup(normalize_values)

    def zrange(self, key: str, start: int, stop: int) -> None:
        self._pipe.zrange(self.with_prefix(key), start, stop)
        self.add_fixup(to_list_str)

    def zcard(self, key: str) -> None:
        self._pipe.zcard(self.with_prefix(key))
        self.add_fixup(int)

    def hset(self, key: str, mapping: dict[str, str]) -> None:
        self._pipe.hset(self.with_prefix(key), mapping=mapping)  # type: ignore
        self.add_fixup(int)

    def hdel(self, key: str, *fields: str) -> None:
        self._pipe.hdel(self.with_prefix(key), *fields)
        self.add_fixup(int)

    def hget(self, key: str, field: str) -> None:
        self._pipe.hget(self.with_prefix(key), field)
        self.add_fixup(to_maybe_str)

    def hmget(self, key: str, *fields: str) -> None:
        self._pipe.hmget(self.with_prefix(key), *fields)
        self.add_fixup(lambda res: {
            field: to_maybe_str(val)
            for field, val in zip(fields, res)
        })

    def hincrby(self, key: str, field: str, inc: float | int) -> None:
        self._pipe.hincrbyfloat(self.with_prefix(key), field, inc)
        self.add_fixup(float)

    def hkeys(self, key: str) -> None:
        self._pipe.hkeys(self.with_prefix(key))
        self.add_fixup(to_list_str)

    def hvals(self, key: str) -> None:
        self._pipe.hvals(self.with_prefix(key))
        self.add_fixup(to_list_str)

    def hgetall(self, key: str) -> None:
        self._pipe.hgetall(self.with_prefix(key))
        self.add_fixup(lambda res: {
            to_maybe_str(field): to_maybe_str(val)
            for field, val in res.items()
        })


class RedisConnection(Runtime[list[str]]):
    """The runtime of the redis backend."""
    def __init__(
            self,
            redis_module: str,
            *,
            cfg: RedisConfig,
            redis_factory: RedisFactory | None = None,
            is_caching_enabled: bool = True,
            verbose_test: bool = True) -> None:
        super().__init__()
        self._conn: RedisWrapper = RedisWrapper(
            cfg=cfg,
            redis_factory=redis_factory,
            is_caching_enabled=is_caching_enabled)
        prefix = cfg.get("prefix", "")
        prefix_str = f"{prefix}:" if prefix else ""
        module = f"{prefix_str}{redis_module}".rstrip(":")
        self._module = f"{module}:" if module else ""
        self._is_print_scripts = verbose_test

    def set_print_scripts(self, is_print_scripts: bool) -> None:
        """
        Whether to print all lua scripts to stdout before registering.

        Args:
            is_print_scripts (bool): Whether to print all lua scripts.
        """
        self._is_print_scripts = is_print_scripts

    @classmethod
    def create_backend(cls) -> LuaBackend:
        return LuaBackend()

    @contextlib.contextmanager
    def pipeline(self) -> Iterator[PipelineAPI]:
        with self.get_connection() as conn:
            with conn.pipeline() as pipe:
                pconn = PipelineConnection(pipe, self._module)
                yield pconn
                if pconn.has_pending():
                    pconn.execute()  # drain pending tasks

    @contextlib.contextmanager
    def get_connection(self) -> Iterator[Redis]:
        """
        Get a raw redis connection.

        Yields:
            Redis: The redis connection.
        """
        with self._conn.get_connection() as conn:
            yield conn

    def get_dynamic_script(self, code: str) -> RedisFunctionBytes:
        """
        Registers a lua script.

        Args:
            code (str): The lua code.

        Returns:
            RedisFunctionBytes: An callable script.
        """
        if is_test() and self._is_print_scripts:
            print(
                "Compiled script:\n-- SCRIPT START --\n"
                f"{code.rstrip()}\n-- SCRIPT END --")
        compute = Script(None, code.encode("utf-8"))
        context = 3

        def get_error(err_msg: str) -> tuple[str, list[str]] | None:
            ustr = "user_script:"
            ix = err_msg.find(ustr)
            if ix < 0:
                return None
            eix = err_msg.find(":", ix + len(ustr))
            if eix < 0:
                return None
            num = int(err_msg[ix + len(ustr):eix])
            rel_line = num

            new_msg = f"{err_msg[:ix + len(ustr)]}{rel_line}{err_msg[eix:]}"
            ctx = [""] * context + code.splitlines()
            return new_msg, ctx[num - 1:num + 2 * context]

        @contextlib.contextmanager
        def get_client(client: Redis | None) -> Iterator[Redis]:
            try:
                if client is None:
                    with self.get_connection() as res:
                        yield res
                else:
                    yield client
            except ResponseError as e:
                handle_err(e)

        def handle_err(exc: ResponseError) -> NoReturn:
            info = None
            if exc.args:
                msg = exc.args[0]
                res = get_error(msg)
                if res is not None:
                    ctx = "\n".join((
                        f"{'>' if ix == context else ' '} {line}"
                        for (ix, line) in enumerate(res[1])))
                    info = f"\nCode:\n{code}\n\nContext:\n{ctx}"
            if info is None:
                info = f": {exc}"
            raise ValueError(f"Error while executing script{info}") from exc

        def execute_bytes_result(
                *,
                keys: list[str],
                args: list[bytes | str | int],
                client: Redis | None) -> bytes:
            with get_client(client) as inner:
                return compute(keys=keys, args=args, client=inner)

        return execute_bytes_result

    def get_prefix(self) -> str:
        """
        The key prefix is added to all keys to create a virtual namespace on
        redis.

        Returns:
            str: The key prefix.
        """
        return self._module

    def with_prefix(self, key: str) -> str:
        """
        Computes the actual key value of this redis instance.

        Args:
            key (str): The key.

        Returns:
            str: The actual key.
        """
        return f"{self.get_prefix()}{key}"

    def get_pubsub_key(self, key: str) -> str:
        """
        Computes the actual key value for a pubsub key.

        Args:
            key (str): The key.

        Returns:
            str: The actual pubsub key.
        """
        return f"{self.get_prefix()}ps:{key}"

    def wait_for(
            self,
            key: str,
            predicate: Callable[[], bool],
            granularity: float = 30.0) -> None:
        """
        Waits until the condition is met.

        Args:
            key (str): The key used for the pubsub channel.

            predicate (Callable[[], bool]): The condition.

            granularity (float, optional): Maximum wait between predicate
            checks in seconds. Defaults to 30.0.
        """
        if predicate():
            return
        with self.get_connection() as conn:
            with conn.pubsub() as psub:
                psub.subscribe(self.get_pubsub_key(key))
                try:
                    while not predicate():
                        psub.get_message(
                            ignore_subscribe_messages=True,
                            timeout=granularity)
                        while psub.get_message() is not None:  # flushing queue
                            pass
                finally:
                    psub.unsubscribe()

    def notify_all(self, key: str) -> None:
        """
        Notifies a pubsub channel.

        Args:
            key (str): The key used for the pubsub channel.
        """
        with self.get_connection() as conn:
            conn.publish(self.get_pubsub_key(key), "notify")

    def ping(self) -> None:
        """
        Pings the redis server (and checks whether the connection is live).
        """
        with self.get_connection() as conn:
            conn.ping()

    def flush_all(self) -> None:
        """
        Removes all keys from the redis database.
        """
        with self.get_connection() as conn:
            conn.flushall()

    def keys_count(self, prefix: str) -> int:
        """
        Counts the number of keys with the given prefix. This method operates
        on raw keys ignoring the `get_prefix()` prefix.

        Args:
            prefix (str): The prefix.

        Returns:
            int: The number of keys.
        """
        full_prefix = f"{prefix}*"
        vals: set[bytes] = set()
        cursor = 0
        count = 10
        with self.get_connection() as conn:
            while True:
                cursor, res = conn.scan(cursor, full_prefix, count)
                vals.update(res)
                if cursor == 0:
                    break
                if count < 4000:
                    count = min(4000, count * 2)
        return len(vals)

    def keys_str(
            self, prefix: str, postfix: str | None = None) -> Iterable[str]:
        """
        Return all keys with the given prefix (and postfix). This method
        operates on raw keys ignoring the `get_prefix()` prefix.

        Args:
            prefix (str): The prefix.

            postfix (str | None, optional): The postfix. Defaults to None.

        Returns:
            Iterable[str]: The raw keys.
        """
        full_prefix = f"{prefix}*{'' if postfix is None else postfix}"
        vals: set[bytes] = set()
        cursor = 0
        count = 10
        with self.get_connection() as conn:
            while True:
                cursor, res = conn.scan(cursor, full_prefix, count)
                vals.update(res)
                if cursor == 0:
                    break
                if count < 1000:
                    count = int(min(1000, count * 1.2))
        return (val.decode("utf-8") for val in vals)

    def prefix_exists(
            self, prefix: str, postfix: str | None = None) -> bool:
        """
        Whether keys exist with the given prefix (and postfix). This method
        operates on raw keys ignoring the `get_prefix()` prefix.

        Args:
            prefix (str): The prefix.

            postfix (str | None, optional): The postfix. Defaults to None.

        Returns:
            bool: Whether there are any keys with the given prefix
            (and postfix).
        """
        full_prefix = f"{prefix}*{'' if postfix is None else postfix}"
        cursor = 0
        count = 10
        with self.get_connection() as conn:
            while True:
                cursor, res = conn.scan(cursor, full_prefix, count)
                if res:
                    return True
                if cursor == 0:
                    return False
                if count < 1000:
                    count = int(min(1000, count * 1.2))

    def exists(self, *keys: str) -> int:
        with self.get_connection() as conn:
            return conn.exists(*(
                self.with_prefix(key) for key in keys))

    def delete(self, *keys: str) -> int:
        with self.get_connection() as conn:
            return conn.delete(*(
                self.with_prefix(key) for key in keys))

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[True],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> str | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode,
            return_previous: Literal[False],
            expire_timestamp: datetime.datetime | None,
            expire_in: float | None,
            keep_ttl: bool) -> bool | None:
        ...

    @overload
    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        ...

    def set(
            self,
            key: str,
            value: str,
            *,
            mode: RSetMode = RSM_ALWAYS,
            return_previous: bool = False,
            expire_timestamp: datetime.datetime | None = None,
            expire_in: float | None = None,
            keep_ttl: bool = False) -> str | bool | None:
        with self.get_connection() as conn:
            expire = None
            if expire_in is not None:
                expire = int(expire_in * 1000.0)
            elif expire_timestamp is not None:
                expire = int(time_diff(now(), expire_timestamp) * 1000.0)
            res = conn.set(
                self.with_prefix(key),
                value,
                get=return_previous,
                nx=(mode == RSM_MISSING),
                xx=(mode == RSM_EXISTS),
                px=expire,
                keepttl=keep_ttl)
            if return_previous:
                if res is not None:
                    return cast(bytes, res).decode("utf-8")
            elif res is None:
                return False
            return res

    def get(self, key: str) -> str | None:
        with self.get_connection() as conn:
            return to_maybe_str(conn.get(self.with_prefix(key)))

    def incrby(self, key: str, inc: float | int) -> float:
        with self.get_connection() as conn:
            return conn.incrbyfloat(self.with_prefix(key), inc)

    def lpush(self, key: str, *values: str) -> int:
        with self.get_connection() as conn:
            return conn.lpush(self.with_prefix(key), *values)

    def rpush(self, key: str, *values: str) -> int:
        with self.get_connection() as conn:
            return conn.rpush(self.with_prefix(key), *values)

    @overload
    def lpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def lpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def lpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.get_connection() as conn:
            res = conn.lpop(self.with_prefix(key), count)
            if count is None:
                return to_maybe_str(res)
            return to_list_str(res)

    @overload
    def rpop(
            self,
            key: str,
            count: None = None) -> str | None:
        ...

    @overload
    def rpop(  # pylint: disable=signature-differs
            self,
            key: str,
            count: int) -> list[str] | None:
        ...

    def rpop(
            self,
            key: str,
            count: int | None = None) -> str | list[str] | None:
        with self.get_connection() as conn:
            res = conn.rpop(self.with_prefix(key), count)
            if count is None:
                return to_maybe_str(res)
            return to_list_str(res)

    def lrange(self, key: str, start: int, stop: int) -> list[str]:
        with self.get_connection() as conn:
            return to_list_str(conn.lrange(self.with_prefix(key), start, stop))

    def llen(self, key: str) -> int:
        with self.get_connection() as conn:
            return conn.llen(self.with_prefix(key))

    def zadd(self, key: str, mapping: dict[str, float]) -> int:
        with self.get_connection() as conn:
            return int(
                conn.zadd(self.with_prefix(key), mapping))  # type: ignore

    def zpop_max(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.get_connection() as conn:
            res = conn.zpopmax(self.with_prefix(key), count)
            return [
                (name.decode("utf-8"), float(score))
                for name, score in res
            ]

    def zpop_min(
            self,
            key: str,
            count: int = 1,
            ) -> list[tuple[str, float]]:
        with self.get_connection() as conn:
            res = conn.zpopmin(self.with_prefix(key), count)
            return [
                (name.decode("utf-8"), float(score))
                for name, score in res
            ]

    def zrange(self, key: str, start: int, stop: int) -> list[str]:
        with self.get_connection() as conn:
            res = conn.zrange(self.with_prefix(key), start, stop)
            return to_list_str(res)

    def zcard(self, key: str) -> int:
        with self.get_connection() as conn:
            return int(conn.zcard(self.with_prefix(key)))

    def hset(self, key: str, mapping: dict[str, str]) -> int:
        with self.get_connection() as conn:
            return conn.hset(
                self.with_prefix(key),
                mapping=mapping)  # type: ignore

    def hdel(self, key: str, *fields: str) -> int:
        with self.get_connection() as conn:
            return conn.hdel(self.with_prefix(key), *fields)

    def hget(self, key: str, field: str) -> str | None:
        with self.get_connection() as conn:
            return to_maybe_str(conn.hget(self.with_prefix(key), field))

    def hmget(self, key: str, *fields: str) -> dict[str, str | None]:
        with self.get_connection() as conn:
            res = conn.hmget(self.with_prefix(key), *fields)
            return {
                field: to_maybe_str(val)
                for field, val in zip(fields, res)
            }

    def hincrby(self, key: str, field: str, inc: float | int) -> float:
        with self.get_connection() as conn:
            return conn.hincrbyfloat(self.with_prefix(key), field, inc)

    def hkeys(self, key: str) -> list[str]:
        with self.get_connection() as conn:
            return to_list_str(conn.hkeys(self.with_prefix(key)))

    def hvals(self, key: str) -> list[str]:
        with self.get_connection() as conn:
            return to_list_str(conn.hvals(self.with_prefix(key)))

    def hgetall(self, key: str) -> dict[str, str]:
        with self.get_connection() as conn:
            return {
                to_maybe_str(field): to_maybe_str(val)
                for field, val in conn.hgetall(self.with_prefix(key)).items()
            }



"""
This module contains various useful functions. The functions in this module are
not necessarily be considered part of the package API and might change or
disappear in the future. Use with caution outside of the package internals.
"""
import datetime
import hashlib
import inspect
import json
import os
import re
import string
import threading
import uuid
from collections.abc import Callable, Iterable
from typing import Any, IO, NoReturn, overload, TypeVar

import pytz


ET = TypeVar('ET')
CT = TypeVar('CT')
RT = TypeVar('RT')
VT = TypeVar('VT')


NL = "\n"


TEST_SALT_LOCK = threading.RLock()
TEST_SALT: dict[str, str] = {}


def is_test() -> bool:
    """
    Whether we are currently running a test with pytest.

    Returns:
        bool: Whether the current execution environment is within a pytest.
    """
    test_id = os.getenv("PYTEST_CURRENT_TEST")
    return test_id is not None


def get_test_salt() -> str | None:
    """
    Creates a unique salt to be used as redis key prefix for unit tests.

    Returns:
        str | None: A unique prefix for the current test. None if we are not
        running inside a pytest environment.
    """
    test_id = os.getenv("PYTEST_CURRENT_TEST")
    if test_id is None:
        return None
    res = TEST_SALT.get(test_id)
    if res is None:
        with TEST_SALT_LOCK:
            res = TEST_SALT.get(test_id)
            if res is None:
                res = f"salt:{uuid.uuid4().hex}"
                TEST_SALT[test_id] = res
    return res


def indent(text: str, amount: int = 0) -> list[str]:
    """
    Breaks a string into lines and indents each line by a set amount.

    Args:
        text (str): The text to break into lines.

        amount (int, optional): The indentation amount. Defaults to 0.

    Returns:
        list[str]: The indented lines.
    """
    istr = " " * amount
    return [
        f"{istr}{line}" for line in text.splitlines()
    ]


def deindent(text: str) -> str:
    """
    Removes a fixed amount of whitespace in front of each line. The amount is
    the maximum length of whitespace that is present on every line. The first
    line and empty lines are treated specially to result in a pleasing output.

    Args:
        text (str): The text to deindent.

    Returns:
        str: The deindented text.
    """
    min_indent = None
    lines = text.rstrip().splitlines()
    if lines and not lines[0]:
        lines.pop(0)
    for line in lines:
        if not line.strip():
            continue
        cur_indent = len(line) - len(line.lstrip())
        if min_indent is None or min_indent > cur_indent:
            min_indent = cur_indent
            if min_indent == 0:
                break
    return "".join((f"{line[min_indent:]}\n" for line in lines))


def lua_fmt(text: str) -> str:
    """
    Formats a multi-line string by deindenting and changing the indent size
    from 4 to 2.

    Args:
        text (str): The multi-line string.

    Returns:
        str: A deindented 2-space indented string.
    """
    return deindent(text).replace("    ", "  ")


def code_fmt(lines: list[str]) -> str:
    """
    Joins a list of lines.

    Args:
        lines (list[str]): The list of lines.

    Returns:
        str: The output as single string.
    """
    return "".join(f"{line.rstrip()}\n" for line in lines)


def get_text_hash(text: str) -> str:
    """
    Computes the blake2b hash of the given string.

    Args:
        text (str): The string.

    Returns:
        str: The hash of length text_hash_size.
    """
    blake = hashlib.blake2b(digest_size=32)
    blake.update(text.encode("utf-8"))
    return blake.hexdigest()


def text_hash_size() -> int:
    """
    The length of the hash produced by get_text_hash.

    Returns:
        int: The length of the hash.
    """
    return 64


def get_short_hash(text: str) -> str:
    """
    Computes a short blake2b hash of the given string.

    Args:
        text (str): The string.

    Returns:
        str: The hash of length short_hash_size.
    """
    blake = hashlib.blake2b(digest_size=4)
    blake.update(text.encode("utf-8"))
    return blake.hexdigest()


def short_hash_size() -> int:
    """
    The length of the hash produced by get_short_hash.

    Returns:
        int: The length of the hash.
    """
    return 8


BUFF_SIZE = 65536  # 64KiB
"""
The buffer size used to compute file hashes.
"""


def get_file_hash(fname: str) -> str:
    """
    Computes the blake2b hash of the given file's content.

    Args:
        fname (str): The filename.

    Returns:
        str: The hash of length file_hash_size.
    """
    blake = hashlib.blake2b(digest_size=32)
    with open(fname, "rb") as fin:
        while True:
            buff = fin.read(BUFF_SIZE)
            if not buff:
                break
            blake.update(buff)
    return blake.hexdigest()


def file_hash_size() -> int:
    """
    The length of the hash produced by get_file_hash.

    Returns:
        int: The length of the hash.
    """
    return 64


def is_hex(text: str) -> bool:
    """
    Whether the string represents a hexadecimal value.

    Args:
        text (str): The string to inspect.

    Returns:
        bool: Whether the string represents a hex value.
    """
    hex_digits = set(string.hexdigits)
    return all(char in hex_digits for char in text)


def only(arr: list[RT]) -> RT:
    """
    Extracts the only item of a single item list.

    Args:
        arr (list[RT]): The single item list.

    Raises:
        ValueError: If the length of the list is not 1.

    Returns:
        RT: The single item.
    """
    if len(arr) != 1:
        raise ValueError(f"array must have exactly one element: {arr}")
    return arr[0]


# time units for logging request durations
ELAPSED_UNITS: list[tuple[int, str]] = [
    (1, "s"),
    (60, "m"),
    (60*60, "h"),
    (60*60*24, "d"),
]
"""Unit conversions for time durations."""


def elapsed_time_string(elapsed: float) -> str:
    """Convert elapsed time into a readable string."""
    cur = ""
    for (conv, unit) in ELAPSED_UNITS:
        if elapsed / conv >= 1 or not cur:
            cur = f"{elapsed / conv:8.3f}{unit}"
        else:
            break
    return cur


def now() -> datetime.datetime:
    """
    Returns a timestamp representing now.

    Returns:
        datetime.datetime: The timestamp representing now.
    """
    return datetime.datetime.now(datetime.timezone.utc).astimezone()


def fmt_time(when: datetime.datetime) -> str:
    """
    Format a timestamp.

    Args:
        when (datetime.datetime): The timestamp.

    Returns:
        str: The formatted time.
    """
    return when.isoformat()


def get_time_str() -> str:
    """
    Formatted now.

    Returns:
        str: Formatted now.
    """
    return fmt_time(now())


def parse_time_str(time_str: str) -> datetime.datetime:
    """
    Parses a timestamp formatted via fmt_time.

    Args:
        time_str (str): The time as string.

    Returns:
        datetime.datetime: A timestamp.
    """
    return datetime.datetime.fromisoformat(time_str)


def time_diff(
        from_time: datetime.datetime, to_time: datetime.datetime) -> float:
    """
    Computes the time difference between time points in seconds.

    Args:
        from_time (datetime.datetime): The earlier time point.
        to_time (datetime.datetime): The later time point.

    Returns:
        float: The time difference in seconds. If `from_time` occurs after
        `to_time` the value will be negative.
    """
    return (to_time - from_time).total_seconds()


def to_bool(value: bool | float | int | str) -> bool:
    """
    Tries converting a given value to a boolean.

    Args:
        value (bool | float | int | str): The value.

    Raises:
        ValueError: If the value cannot be interpreted as boolean.

    Returns:
        bool: The value interpreted as boolean. A string can be 'true' or
        'false' and numbers are interpreted as True if they're non-zero.
    """
    value = f"{value}".lower()
    if value == "true":
        return True
    if value == "false":
        return False
    try:
        return bool(int(float(value)))
    except ValueError:
        pass
    raise ValueError(f"{value} cannot be interpreted as bool")


def to_list(value: Any) -> list[Any]:
    """
    Ensures a given value is a list.

    Args:
        value (Any): The value.

    Raises:
        ValueError: If it is not a list.

    Returns:
        list[Any]: The value as list.
    """
    if not isinstance(value, list):
        raise ValueError(f"{value} is not a list")
    return value


def is_int(value: Any) -> bool:
    """
    Determines whether a value can be interpreted as integer.

    Args:
        value (Any): The value.

    Returns:
        bool: Whether the value can be interpreted as integer.
    """
    try:
        int(value)
        return True
    except ValueError:
        return False


def is_float(value: Any) -> bool:
    """
    Determines whether a value can be interpreted as float.

    Args:
        value (Any): The value.

    Returns:
        bool: Whether the value can be interpreted as float.
    """
    try:
        float(value)
        return True
    except (ValueError, TypeError):
        return False


def to_number_str(value: int | float) -> str:
    """
    Converts a number into a string. If the number appears as int it will be
    printed without decimal places even if its type is float.

    Args:
        value (int | float): The number.

    Returns:
        str: The string representation.
    """
    if int(value) == value:
        return f"{int(value)}"
    return f"{value}"


def is_json(value: str) -> bool:
    """
    Determines whether a value can be decoded as JSON.

    Args:
        value (Any): The value.

    Returns:
        bool: Whether the value can be decoded as JSON.
    """
    try:
        json.loads(value)
    except json.JSONDecodeError:
        return False
    return True


def report_json_error(err: json.JSONDecodeError) -> NoReturn:
    """
    Provides information about JSON decode errors.

    Args:
        err (json.JSONDecodeError): The JSON decode error.

    Raises:
        ValueError: The error raised again with more information about the
        issue.
    """
    raise ValueError(
        f"JSON parse error ({err.lineno}:{err.colno}): "
        f"{repr(err.doc)}") from err


def json_maybe_read(data: str) -> Any | None:
    """
    Maybe read a JSON.

    Args:
        data (str): The data that might contain a JSON.

    Returns:
        Any | None: The JSON or None if it is not a JSON decodable input.
    """
    try:
        return json.loads(data)
    except json.JSONDecodeError:
        return None


def json_load(fin: IO[str]) -> Any:
    """
    Load a JSON document from a file-like object.

    Args:
        fin (IO[str]): The file-like input.

    Raises:
        e (ValueError): If the document cannot be parsed as JSON.

    Returns:
        Any: The JSON.
    """
    try:
        return json.load(fin)
    except json.JSONDecodeError as e:
        report_json_error(e)


def json_dump(obj: Any, fout: IO[str]) -> None:
    """
    Writes the given object as JSON to the given file-like object.

    Args:
        obj (Any): The object to dump as JSON.
        fout (IO[str]): The file-like output.
    """
    print(json_pretty(obj), file=fout)


def json_pretty(obj: Any) -> str:
    """
    Dumps the given object as pretty JSON string.

    Args:
        obj (Any): The object to dump.

    Returns:
        str: The pretty JSON string.
    """
    return json.dumps(obj, sort_keys=True, indent=2)


def json_compact(obj: Any) -> bytes:
    """
    Dumps the given object as compact JSON bytes.

    Args:
        obj (Any): The object to dump.

    Returns:
        bytes: The compact JSON bytes.
    """
    return json.dumps(
        obj,
        sort_keys=True,
        indent=None,
        separators=(",", ":")).encode("utf-8")


def json_read(data: bytes) -> Any:
    """
    Parses the given bytes as JSON.

    Args:
        data (bytes): The byte input.

    Raises:
        e (ValueError): If the document cannot be parsed as JSON.

    Returns:
        Any: The JSON.
    """
    try:
        return json.loads(data.decode("utf-8"))
    except json.JSONDecodeError as e:
        report_json_error(e)


def read_jsonl(fin: IO[str]) -> Iterable[Any]:
    """
    Reads a JSONL from a file-like object. Each line in the file is interpreted
    as JSON. Each JSON cannot contain a newline character.

    Args:
        fin (IO[str]): The file-like input.

    Raises:
        e (ValueError): If any line cannot be parsed as JSON.

    Yields:
        Any: One JSON object for each line in the input.
    """
    for line in fin:
        line = line.rstrip()
        if not line:
            continue
        try:
            yield json.loads(line)
        except json.JSONDecodeError as e:
            report_json_error(e)


def from_timestamp(timestamp: float) -> datetime.datetime:
    """
    Converts a float POSIX-timestamp to a datetime object.

    Args:
        timestamp (float): The float POSIX-timestamp.

    Returns:
        datetime.datetime: The datetime object.
    """
    return datetime.datetime.fromtimestamp(timestamp, pytz.utc)


def to_timestamp(time: datetime.datetime) -> float:
    """
    Converts a datetime object into a float POSIX-timestamp.

    Args:
        time (datetime.datetime): The datetime object.

    Returns:
        float: The float POSIX-timestamp.
    """
    return time.timestamp()


def now_ts() -> datetime.datetime:
    """
    Returns a datetime object representing the current time.

    Returns:
        datetime.datetime: The datetime object.
    """
    return datetime.datetime.now(pytz.utc)


def get_function_info(*, clazz: type) -> tuple[str, int, str]:
    """
    Computes where in the current execution stack the instruction pointer was
    in the given class.

    Args:
        clazz (type): The class of interest.

    Returns:
        tuple[str, int, str]: At which file, line, and function the execution
        is when walking down the stack.
    """
    stack = inspect.stack()

    def get_method(cur_clazz: type) -> tuple[str, int, str] | None:
        class_filename = inspect.getfile(cur_clazz)
        for level in stack:
            if os.path.samefile(level.filename, class_filename):
                return level.filename, level.lineno, level.function
        return None

    queue = [clazz]
    while queue:
        cur = queue.pop(0)
        res = get_method(cur)
        if res is not None:
            return res
        queue.extend(cur.__bases__)
    frame = stack[1]
    return frame.filename, frame.lineno, frame.function


def get_relative_function_info(
        depth: int) -> tuple[str, int, str, dict[str, Any]]:
    """
    Computes the execution position of the stack frame `depth` levels down.

    Args:
        depth (int): The depth of the stack to inspect.

    Returns:
        tuple[str, int, str, dict[str, Any]]: The filename, line number,
        function name, and local variables.
    """
    depth += 1
    stack = inspect.stack()
    if depth >= len(stack):
        return "unknown", -1, "unknown", {}
    frame = stack[depth]
    return frame.filename, frame.lineno, frame.function, frame.frame.f_locals


def identity(obj: RT) -> RT:
    """
    A generic identity function.

    Args:
        obj (RT): The object.

    Returns:
        RT: The same object.
    """
    return obj


NUMBER_PATTERN = re.compile(r"\d+")
"""Regex to extract numbers."""


def extract_list(
        arr: Iterable[str],
        prefix: str | None = None,
        postfix: str | None = None) -> Iterable[tuple[str, str]]:
    """
    Extract substrings from a stream of strings.

    Args:
        arr (Iterable[str]): The stream of input strings.

        prefix (str | None, optional): The prefix to match. Defaults to None.

        postfix (str | None, optional): The postfix to match. Defaults to None.

    Yields:
        tuple[str, str]: Returns each matching string.
        The first element of the tuple is the full string. The second element
        is only the substring after removing the matching prefix and postfix.
    """
    for elem in arr:
        text = elem
        if prefix is not None:
            if not text.startswith(prefix):
                continue
            text = text[len(prefix):]
        if postfix is not None:
            if not text.endswith(postfix):
                continue
            text = text[:-len(postfix)]
        yield (elem, text)


def extract_number(
        arr: Iterable[str],
        prefix: str | None = None,
        postfix: str | None = None) -> Iterable[tuple[str, int]]:
    """
    Extract numbers from a stream of strings. The numbers are extracted from
    the substring excluding the prefix and postfix (if set). Additional
    non-numeric characters are allowed.

    Args:
        arr (Iterable[str]): The stream of strings.

        prefix (str | None, optional): The prefix must match. Defaults to None.

        postfix (str | None, optional): The postfix must match.
        Defaults to None.

    Yields:
        tuple[str, int]: The matching strings and the extracted
        numbers. The first element of the tuple is the full string. The second
        element is the extracted number.
    """

    def get_num(text: str) -> int | None:
        match = re.search(NUMBER_PATTERN, text)
        if match is None:
            return None
        try:
            return int(match.group())
        except ValueError:
            return None

    for elem, text in extract_list(arr, prefix=prefix, postfix=postfix):
        num = get_num(text)
        if num is None:
            continue
        yield elem, num


def highest_number(
        arr: Iterable[str],
        prefix: str | None = None,
        postfix: str | None = None) -> tuple[str, int] | None:
    """
    Retrieves the highest number extracted from a stream of strings. The
    numbers are extracted from the substring excluding the prefix and postfix
    (if set). Additional non-numeric characters are allowed. Only the match
    with the highest number is returned.

    Args:
        arr (Iterable[str]): The stream of strings.

        prefix (str | None, optional): The prefix to match. Defaults to None.

        postfix (str | None, optional): The postfix to match. Defaults to None.

    Returns:
        tuple[str, int] | None: The matching element with the highest number.
        The first element of the tuple is the matching string. The second
        element is the extracted number.
    """
    res = None
    res_num = 0
    for elem, num in extract_number(arr, prefix=prefix, postfix=postfix):
        if res is None or num > res_num:
            res = elem
            res_num = num
    return None if res is None else (res, res_num)


def retain_some(
        arr: Iterable[VT],
        count: int,
        *,
        key: Callable[[VT], Any],
        reverse: bool = False,
        keep_last: bool = True) -> tuple[list[VT], list[VT]]:
    """
    Retains up to `count` elements from `arr`.

    Args:
        arr (Iterable[VT]): The stream of elements.

        count (int): The number of elements to retain.

        key (Callable[[VT], Any]): The key for elements to define an order.

        reverse (bool, optional): Whether to reverse the order. Defaults to
        False.

        keep_last (bool, optional): Whether to retain the last
        elements. Defaults to True.

    Returns:
        tuple[list[VT], list[VT]]: The first element of the tuple is the list
        of elements to retain. The second element is the list of elements to
        remove.
    """
    res: list[VT] = []
    to_delete: list[VT] = []
    if keep_last:
        for elem in arr:
            if len(res) <= count:
                res.append(elem)
                continue
            res.sort(key=key, reverse=reverse)
            to_delete.extend(res[:-count])
            res = res[-count:]
            res.append(elem)
    else:
        for elem in arr:
            res.append(elem)
            if len(res) < count:
                continue
            res.sort(key=key, reverse=reverse)
            to_delete.extend(res[:-count])
            res = res[-count:]
    res.sort(key=key, reverse=reverse)
    return res, to_delete


def python_module() -> str:
    """
    Computes the caller's python module.

    Raises:
        ValueError: If the module cannot be found.

    Returns:
        str: The full name of the module.
    """
    stack = inspect.stack()
    module = inspect.getmodule(stack[1][0])
    if module is None:
        raise ValueError("module not found")
    res = module.__name__
    if res != "__main__":
        return res
    package = module.__package__
    if package is None:
        package = ""
    mfname = module.__file__
    if mfname is None:
        return package
    fname = os.path.basename(mfname)
    fname = fname.removesuffix(".py")
    if fname in ("__init__", "__main__"):
        return package
    return f"{package}.{fname}"


def parent_python_module(p_module: str) -> str:
    """
    Computes the parent module of a given module.

    Args:
        p_module (str): The module.

    Returns:
        str: The parent module or `""` if the module was already a top level
        module.
    """
    dot_ix = p_module.rfind(".")
    if dot_ix < 0:
        return ""
    return p_module[:dot_ix]


def check_pid_exists(pid: int) -> bool:
    """
    Checks whether a given pid exists.

    Args:
        pid (int): The process id.

    Returns:
        bool: Whether a process with the given id exists.
    """
    try:
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def ideal_thread_count() -> int:
    """
    Computes the ideal thread count for the given machine.

    Returns:
        int: The ideal number of threads.
    """
    res = os.cpu_count()
    if res is None:
        return 4
    return res


def escape(text: str, subs: dict[str, str]) -> str:
    """
    Escapes characters in the text according to the substitution dictionary.

    Args:
        text (str): The text.

        subs (dict[str, str]): The substitution dictionary. For example,
        `{"\\n", "n"}` would replace all newlines with `\\n` in the text.

    Returns:
        str: The escaped text.
    """
    text = text.replace("\\", "\\\\")
    for key, repl in subs.items():
        text = text.replace(key, f"\\{repl}")
    return text


def unescape(text: str, subs: dict[str, str]) -> str:
    """
    Unescapes characters in the text according to the substitution dictionary.

    Args:
        text (str): The text.

        subs (dict[str, str]): The substitution dictionary. For example,
        `{"n": "\\n"}` would replace all `\\n` sequences in the text with
        newlines.

    Returns:
        str: The unescaped text.
    """
    res: list[str] = []
    in_escape = False
    for c in text:
        if in_escape:
            in_escape = False
            if c == "\\":
                res.append("\\")
                continue
            done = False
            for key, repl in subs.items():
                if c == key:
                    res.append(repl)
                    done = True
                    break
            if done:
                continue
        if c == "\\":
            in_escape = True
            continue
        res.append(c)
    return "".join(res)


@overload
def to_maybe_str(res: bytes) -> str:
    ...


@overload
def to_maybe_str(res: None) -> None:
    ...


def to_maybe_str(res: bytes | None) -> str | None:
    """
    Converts bytes to a string. If the input is None then only None will be
    returned.

    Args:
        res (bytes | None): The bytes.

    Returns:
        str | None: The string from the bytes or None if the input was None.
    """
    if res is None:
        return res
    return res.decode("utf-8")


@overload
def to_list_str(res: Iterable[bytes]) -> list[str]:
    ...


@overload
def to_list_str(res: None) -> None:
    ...


def to_list_str(res: Iterable[bytes] | None) -> list[str] | None:
    """
    Converts a list of bytes into a list of strings. If the input is None
    then only None is returned.

    Args:
        res (Iterable[bytes] | None): The list of bytes.

    Returns:
        list[str] | None: The list of strings. If the input was None then None
        is returned.
    """
    if res is None:
        return res
    return [val.decode("utf-8") for val in res]


def normalize_values(res: Any) -> Any:
    """
    Converts all bytes into string in the data structure.

    Args:
        res (Any): The data structure. Can be plain bytes, lists, tuples,
        dictionaries, and other literal values.

    Returns:
        Any: Returns the data structure with all bytes converted to strings.
    """
    if res is None:
        return None
    if isinstance(res, bytes):
        return res.decode("utf-8")
    if isinstance(res, list):
        return [normalize_values(val) for val in res]
    if isinstance(res, tuple):
        return tuple(normalize_values(val) for val in res)
    if isinstance(res, dict):
        return {
            normalize_values(key): normalize_values(value)
            for key, value in res.items()
        }
    return res



"""This module defines the base class for redipy backends. The backend defined
here handles script compilation. For redis functionality look at the runtime.
"""
from typing import Generic, Protocol, TYPE_CHECKING, TypeVar

from redipy.graph.cmd import CommandObj
from redipy.graph.expr import ExprObj, JSONType
from redipy.graph.seq import SequenceObj


if TYPE_CHECKING:
    from redipy.api import PipelineAPI, RedisAPI
    from redipy.backend.runtime import Runtime


T = TypeVar('T')
ET = TypeVar('ET')
CC = TypeVar('CC')
EC = TypeVar('EC')
R = TypeVar('R', bound='Runtime')


class ExecFunction(Protocol):  # pylint: disable=too-few-public-methods
    """
    A redis script. Calling this function executes the script.
    """
    def __call__(
            self,
            *,
            keys: dict[str, str],
            args: dict[str, JSONType],
            client: 'RedisAPI | PipelineAPI | None' = None) -> JSONType:
        """
        Executes the redis script.

        Args:
            keys (dict[str, str]): A dictionary of keys registered in the
            script. It is common convention to pass keys used in the script via
            this parameter but it is not necessary to do so. Importantly, keys
            can also be generated inside the script. The keys of this
            dictionary are the names of the redis keys as registered in the
            script.

            args (dict[str, JSONType]): Arguments to the script. The keys of
            this dictionary are the names of the arguments as registered in the
            script.

            client (RedisAPI | PipelineAPI | None): An optionally different
            execution environment. Note, that the different execution
            environment has to be of the same runtime type. Setting the client
            allows scripts to execute inside a pipeline. Defaults to None.

        Returns:
            JSONType: The result of the script.
        """
        raise NotImplementedError()


class Backend(Generic[T, ET, CC, EC, R]):
    """
    This class handles compilation of redis scripts.
    """
    def translate(self, seq: SequenceObj) -> T:
        """
        Translate a script into the backends internal type.

        Args:
            seq (SequenceObj): The root sequence object.

        Returns:
            T: The internal representation of the script.
        """
        ctx = self.create_command_context()
        res = self.compile_sequence(ctx, seq)
        return self.finish(ctx, res)

    def create_command_context(self) -> CC:
        """
        Creates a script context.

        Returns:
            CC: The newly created script context.
        """
        raise NotImplementedError()

    def finish(self, ctx: CC, script: T) -> T:
        """
        Finalize the generated script.

        Args:
            ctx (CC): The script context.
            script (T): The internal representation of the script.

        Returns:
            T: The final internal representation of the script.
        """
        raise NotImplementedError()

    def compile_sequence(self, ctx: CC, seq: SequenceObj) -> T:
        """
        Compiles a sequence object.

        Args:
            ctx (CC): The script context.
            seq (SequenceObj): The sequence object.

        Returns:
            T: The internal representation of the sequence object.
        """
        raise NotImplementedError()

    def compile_command(self, ctx: CC, cmd: CommandObj) -> T:
        """
        Compiles a statemement.

        Args:
            ctx (CC): The script context.
            cmd (CommandObj): The statement object.

        Returns:
            T: The internal representation of the statement.
        """
        raise NotImplementedError()

    def compile_expr(self, ctx: EC, expr: ExprObj) -> ET:
        """
        Compiles an expression.

        Args:
            ctx (EC): The script context.
            expr (ExprObj): The expression object.

        Returns:
            ET: The internal representation of the statement.
        """
        raise NotImplementedError()

    def create_executable(
            self,
            code: T,
            runtime: R) -> ExecFunction:
        """
        Create an executable function from an internal representation of the
        script.

        Args:
            code (T): The internal representation of the script.
            runtime (R): The associated runtime.

        Returns:
            ExecFunction: A function that executes the script when called.
        """
        raise NotImplementedError()



"""Tests redis functionality via the main API class."""
from collections.abc import Callable
from test.util import get_test_config
from typing import cast

import pytest

from redipy.api import PipelineAPI
from redipy.graph.expr import JSONType
from redipy.main import Redis
from redipy.symbolic.core import KeyVariable
from redipy.symbolic.expr import Expr
from redipy.symbolic.rhash import RedisHash
from redipy.symbolic.rlist import RedisList
from redipy.symbolic.rvar import RedisVar
from redipy.symbolic.rzset import RedisSortedSet
from redipy.symbolic.seq import FnContext
from redipy.util import code_fmt, lua_fmt


@pytest.mark.parametrize("rt_lua", [False, True])
def test_api(rt_lua: bool) -> None:
    """
    Tests redis functionality via the main API class.

    Args:
        rt_lua (bool): Whether to use the redis or memory runtime.
    """
    # pylint: disable=unnecessary-lambda
    code_name = None
    lua_script = None

    def set_lua_script(name: str | None, lscript: str | None) -> None:
        nonlocal code_name
        nonlocal lua_script

        code_name = name
        lua_script = lscript

    def code_hook(code: list[str]) -> None:
        nonlocal code_name
        nonlocal lua_script

        if lua_script is None:
            return
        filtered_code = []
        in_helpers = False
        for line in code:
            if line == "-- HELPERS START --":
                in_helpers = True
                continue
            if line == "-- HELPERS END --":
                in_helpers = False
                continue
            if in_helpers:
                continue
            filtered_code.append(line)
        code_str = code_fmt(filtered_code)
        lscript = lua_fmt(lua_script)
        success = False
        try:
            assert code_str == lscript
            success = True
        finally:
            if not success:
                print(f"script name: {code_name}")

    redis = Redis(
        cfg=get_test_config() if rt_lua else None,
        lua_code_hook=code_hook)

    def check(
            name: str,
            *,
            setup: Callable[[str], JSONType],
            normal: Callable[[str], JSONType],
            setup_pipe: Callable[[PipelineAPI, str], None],
            pipeline: Callable[[PipelineAPI, str], None],
            lua: Callable[[FnContext, KeyVariable], Expr],
            code: str,
            teardown: Callable[[str], JSONType],
            output_setup: JSONType,
            output: JSONType,
            output_teardown: JSONType,
            lua_patch: Callable[[JSONType], JSONType] | None = None) -> None:
        print(f"testing {name}")
        key = "foo"

        def lua_patch_id(res: JSONType) -> JSONType:
            return res

        if lua_patch is None:
            lua_patch = lua_patch_id

        assert setup(key) == output_setup
        result = normal(key)
        assert result == output
        assert teardown(key) == output_teardown

        with redis.pipeline() as pipe:
            assert setup_pipe(pipe, key) is None
            assert pipeline(pipe, key) is None
            setup_result, result = pipe.execute()
        assert setup_result == output_setup
        assert result == output
        assert teardown(key) == output_teardown

        ctx = FnContext()
        key_var = ctx.add_key("key")
        lcl = ctx.add_local(lua(ctx, key_var))
        ctx.set_return_value(lcl)

        lua_code = lua_fmt(f"""
            --[[ KEYV
            key
            ]]
            --[[ ARGV
            ]]
            local key_0 = (KEYS[1])  -- key
            local var_0 = {code}
            return cjson.encode(var_0)
        """)
        set_lua_script(name, lua_code)
        fun = redis.register_script(ctx)

        assert setup(key) == output_setup
        result = lua_patch(fun(keys={"key": key}, args={}))
        assert result == output
        assert teardown(key) == output_teardown

        assert setup(key) == output_setup
        result = lua_patch(fun(keys={"key": key}, args={}, client=redis))
        assert result == output
        assert teardown(key) == output_teardown

        assert setup(key) == output_setup
        result = lua_patch(
            fun(keys={"key": key}, args={}, client=redis.get_runtime()))
        assert result == output
        assert teardown(key) == output_teardown

        with redis.pipeline() as pipe:
            assert setup_pipe(pipe, key) is None
            assert fun(keys={"key": key}, args={}, client=pipe) is None
            setup_result, result = pipe.execute()
        assert setup_result == output_setup
        assert lua_patch(result) == output
        assert teardown(key) == output_teardown

    check(
        "exists",
        setup=lambda key: redis.set(key, "a"),
        normal=lambda key: redis.exists(key),
        setup_pipe=lambda pipe, key: pipe.set(key, "a"),
        pipeline=lambda pipe, key: pipe.exists(key),
        lua=lambda ctx, key: RedisVar(key).exists(),
        code="redis.call(\"exists\", key_0)",
        teardown=lambda key: [redis.get(key), redis.delete(key)],
        output_setup=True,
        output=1,
        output_teardown=["a", 1])

    check(
        "incrby",
        setup=lambda key: redis.set(key, "0.25"),
        normal=lambda key: redis.incrby(key, 0.5),
        setup_pipe=lambda pipe, key: pipe.set(key, "0.25"),
        pipeline=lambda pipe, key: pipe.incrby(key, 0.5),
        lua=lambda ctx, key: RedisVar(key).incrby(0.5),
        code="tonumber(redis.call(\"incrbyfloat\", key_0, 0.5))",
        teardown=lambda key: [redis.get(key), redis.delete(key)],
        output_setup=True,
        output=0.75,
        output_teardown=["0.75", 1])

    check(
        "lpop_0",
        setup=lambda key: redis.lpush(key, "a"),
        normal=lambda key: redis.lpop(key),
        setup_pipe=lambda pipe, key: pipe.lpush(key, "a"),
        pipeline=lambda pipe, key: pipe.lpop(key),
        lua=lambda ctx, key: RedisList(key).lpop(),
        code="(redis.call(\"lpop\", key_0) or nil)",
        # use exists here and llen below
        teardown=lambda key: redis.exists(key),
        output_setup=1,
        output="a",
        output_teardown=0)

    check(
        "lpop_1",
        setup=lambda key: redis.lpush(key, "a", "b", "c"),
        normal=lambda key: redis.lpop(key, 2),
        setup_pipe=lambda pipe, key: pipe.lpush(key, "a", "b", "c"),
        pipeline=lambda pipe, key: pipe.lpop(key, 2),
        lua=lambda ctx, key: RedisList(key).lpop(2),
        code="redis.call(\"lpop\", key_0, 2)",
        teardown=lambda key: [
            redis.llen(key), redis.delete(key), redis.exists(key)],
        output_setup=3,
        output=["c", "b"],
        output_teardown=[1, 1, 0])

    check(
        "rpop_0",
        setup=lambda key: redis.rpush(key, "a"),
        normal=lambda key: redis.rpop(key),
        setup_pipe=lambda pipe, key: pipe.rpush(key, "a"),
        pipeline=lambda pipe, key: pipe.rpop(key),
        lua=lambda ctx, key: RedisList(key).rpop(),
        code="(redis.call(\"rpop\", key_0) or nil)",
        teardown=lambda key: redis.exists(key),
        output_setup=1,
        output="a",
        output_teardown=0)

    check(
        "rpop_1",
        setup=lambda key: redis.rpush(key, "a", "b", "c"),
        normal=lambda key: redis.rpop(key, 2),
        setup_pipe=lambda pipe, key: pipe.rpush(key, "a", "b", "c"),
        pipeline=lambda pipe, key: pipe.rpop(key, 2),
        lua=lambda ctx, key: RedisList(key).rpop(2),
        code="redis.call(\"rpop\", key_0, 2)",
        teardown=lambda key: [
            redis.llen(key), redis.delete(key), redis.exists(key)],
        output_setup=3,
        output=["c", "b"],
        output_teardown=[1, 1, 0])

    check(
        "zpopmax_0",
        setup=lambda key: redis.zadd(key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        normal=lambda key: redis.zpop_max(key, 2),
        setup_pipe=lambda pipe, key: pipe.zadd(
            key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        pipeline=lambda pipe, key: pipe.zpop_max(key, 2),
        lua=lambda ctx, key: RedisSortedSet(key).pop_max(2),
        code="redipy.pairlist_scores(redis.call(\"zpopmax\", key_0, 2))",
        lua_patch=lambda res: [tuple(elem) for elem in cast(list, res)],
        teardown=lambda key: [
            redis.zcard(key), redis.delete(key), redis.zcard(key)],
        output_setup=3,
        output=[("c", 0.75), ("b", 0.5)],
        output_teardown=[1, True, 0])

    check(
        "zpopmax_1",
        setup=lambda key: redis.zadd(key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        normal=lambda key: redis.zpop_max(key),
        setup_pipe=lambda pipe, key: pipe.zadd(
            key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        pipeline=lambda pipe, key: pipe.zpop_max(key),
        lua=lambda ctx, key: RedisSortedSet(key).pop_max(),
        code="redipy.pairlist_scores(redis.call(\"zpopmax\", key_0))",
        lua_patch=lambda res: [tuple(elem) for elem in cast(list, res)],
        teardown=lambda key: [
            redis.zcard(key), redis.delete(key), redis.zcard(key)],
        output_setup=3,
        output=[("c", 0.75)],
        output_teardown=[2, True, 0])

    check(
        "zpopmin_0",
        setup=lambda key: redis.zadd(key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        normal=lambda key: redis.zpop_min(key, 2),
        setup_pipe=lambda pipe, key: pipe.zadd(
            key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        pipeline=lambda pipe, key: pipe.zpop_min(key, 2),
        lua=lambda ctx, key: RedisSortedSet(key).pop_min(2),
        code="redipy.pairlist_scores(redis.call(\"zpopmin\", key_0, 2))",
        lua_patch=lambda res: [tuple(elem) for elem in cast(list, res)],
        teardown=lambda key: [
            redis.zcard(key), redis.delete(key), redis.zcard(key)],
        output_setup=3,
        output=[("a", 0.25), ("b", 0.5)],
        output_teardown=[1, True, 0])

    check(
        "zpopmin_1",
        setup=lambda key: redis.zadd(key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        normal=lambda key: redis.zpop_min(key),
        setup_pipe=lambda pipe, key: pipe.zadd(
            key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        pipeline=lambda pipe, key: pipe.zpop_min(key),
        lua=lambda ctx, key: RedisSortedSet(key).pop_min(),
        code="redipy.pairlist_scores(redis.call(\"zpopmin\", key_0))",
        lua_patch=lambda res: [tuple(elem) for elem in cast(list, res)],
        teardown=lambda key: [
            redis.zcard(key), redis.delete(key), redis.zcard(key)],
        output_setup=3,
        output=[("a", 0.25)],
        output_teardown=[2, True, 0])

    check(
        "zrange_0",
        setup=lambda key: redis.zadd(key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        normal=lambda key: redis.zrange(key, 1, 2),
        setup_pipe=lambda pipe, key: pipe.zadd(
            key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        pipeline=lambda pipe, key: pipe.zrange(key, 1, 2),
        lua=lambda ctx, key: RedisSortedSet(key).range(1, 2),
        code="redis.call(\"zrange\", key_0, 1, 2)",
        teardown=lambda key: [redis.delete(key), redis.zcard(key)],
        output_setup=3,
        output=["b", "c"],
        output_teardown=[True, 0])

    check(
        "zrange_1",
        setup=lambda key: redis.zadd(key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        normal=lambda key: redis.zrange(key, 0, -2),
        setup_pipe=lambda pipe, key: pipe.zadd(
            key, {"a": 0.25, "b": 0.5, "c": 0.75}),
        pipeline=lambda pipe, key: pipe.zrange(key, 0, -2),
        lua=lambda ctx, key: RedisSortedSet(key).range(0, -2),
        code="redis.call(\"zrange\", key_0, 0, -2)",
        teardown=lambda key: [redis.delete(key), redis.zcard(key)],
        output_setup=3,
        output=["a", "b"],
        output_teardown=[True, 0])

    check(
        "hget",
        setup=lambda key: redis.hset(key, {"a": "0", "b": "1", "c": "2"}),
        normal=lambda key: redis.hget(key, "b"),
        setup_pipe=lambda pipe, key: pipe.hset(
            key, {"a": "0", "b": "1", "c": "2"}),
        pipeline=lambda pipe, key: pipe.hget(key, "b"),
        lua=lambda ctx, key: RedisHash(key).hget("b"),
        code="(redis.call(\"hget\", key_0, \"b\") or nil)",
        teardown=lambda key: [
            redis.exists(key),
            redis.hdel(key, "a"),
            redis.exists(key),
            redis.hdel(key, "b", "c"),
            redis.exists(key),
        ],
        output_setup=3,
        output="1",
        output_teardown=[1, 1, 1, 2, 0])

    check(
        "hmget",
        setup=lambda key: redis.hset(key, {"d": "3", "e": "4", "f": "5"}),
        normal=lambda key: redis.hmget(key, "b", "c", "d", "e"),
        setup_pipe=lambda pipe, key: pipe.hset(
            key, {"d": "3", "e": "4", "f": "5"}),
        pipeline=lambda pipe, key: pipe.hmget(key, "b", "c", "d", "e"),
        lua=lambda ctx, key: RedisHash(key).hmget("b", "c", "d", "e"),
        code=(
            "redipy.keyval_dict(redis.call(\"hmget\", "
            "key_0, \"b\", \"c\", \"d\", \"e\"), \"b\", \"c\", \"d\", \"e\")"
        ),
        teardown=lambda key: [
            redis.exists(key),
            redis.delete(key),
            redis.hgetall(key),
        ],
        output_setup=3,
        output={"b": None, "c": None, "d": "3", "e": "4"},
        output_teardown=[1, 1, {}])

    check(
        "hincrby",
        setup=lambda key: redis.hset(key, {"d": "3", "e": "4", "f": "5"}),
        normal=lambda key: redis.hincrby(key, "e", 2),
        setup_pipe=lambda pipe, key: pipe.hset(
            key, {"d": "3", "e": "4", "f": "5"}),
        pipeline=lambda pipe, key: pipe.hincrby(key, "e", 2),
        lua=lambda ctx, key: RedisHash(key).hincrby("e", 2),
        code="tonumber(redis.call(\"hincrbyfloat\", key_0, \"e\", 2))",
        teardown=lambda key: [
            redis.hgetall(key),
            redis.delete(key),
            redis.exists(key),
        ],
        output_setup=3,
        output=6,
        output_teardown=[{"d": "3", "e": "6", "f": "5"}, 1, 0])

    check(
        "hdel",
        setup=lambda key: redis.hset(key, {"d": "3", "e": "4", "f": "5"}),
        normal=lambda key: redis.hdel(key, "c", "d", "e"),
        setup_pipe=lambda pipe, key: pipe.hset(
            key, {"d": "3", "e": "4", "f": "5"}),
        pipeline=lambda pipe, key: pipe.hdel(key, "c", "d", "e"),
        lua=lambda ctx, key: RedisHash(key).hdel("c", "d", "e"),
        code="redis.call(\"hdel\", key_0, \"c\", \"d\", \"e\")",
        teardown=lambda key: [redis.hgetall(key), redis.delete(key)],
        output_setup=3,
        output=2,
        output_teardown=[{"f": "5"}, 1])

    check(
        "hkeys",
        setup=lambda key: redis.hset(key, {"d": "3", "e": "4", "f": "5"}),
        normal=lambda key: redis.hkeys(key),
        setup_pipe=lambda pipe, key: pipe.hset(
            key, {"d": "3", "e": "4", "f": "5"}),
        pipeline=lambda pipe, key: pipe.hkeys(key),
        lua=lambda ctx, key: RedisHash(key).hkeys(),
        code="redis.call(\"hkeys\", key_0)",
        teardown=lambda key: redis.delete(key),
        output_setup=3,
        output=["d", "e", "f"],
        output_teardown=1)

    check(
        "hvals",
        setup=lambda key: redis.hset(key, {"d": "3", "e": "4", "f": "5"}),
        normal=lambda key: redis.hvals(key),
        setup_pipe=lambda pipe, key: pipe.hset(
            key, {"d": "3", "e": "4", "f": "5"}),
        pipeline=lambda pipe, key: pipe.hvals(key),
        lua=lambda ctx, key: RedisHash(key).hvals(),
        code="redis.call(\"hvals\", key_0)",
        teardown=lambda key: redis.delete(key),
        output_setup=3,
        output=["3", "4", "5"],
        output_teardown=1)

    check(
        "hgetall",
        setup=lambda key: redis.hset(key, {"d": "3", "e": "4", "f": "5"}),
        normal=lambda key: redis.hgetall(key),
        setup_pipe=lambda pipe, key: pipe.hset(
            key, {"d": "3", "e": "4", "f": "5"}),
        pipeline=lambda pipe, key: pipe.hgetall(key),
        lua=lambda ctx, key: RedisHash(key).hgetall(),
        code="redipy.pairlist_dict(redis.call(\"hgetall\", key_0))",
        teardown=lambda key: redis.delete(key),
        output_setup=3,
        output={"d": "3", "e": "4", "f": "5"},
        output_teardown=1)

    check(
        "lrange_0",
        setup=lambda key: redis.rpush(key, "a", "b", "c"),
        normal=lambda key: redis.lrange(key, 0, 0),
        setup_pipe=lambda pipe, key: pipe.rpush(key, "a", "b", "c"),
        pipeline=lambda pipe, key: pipe.lrange(key, 0, 0),
        lua=lambda ctx, key: RedisList(key).lrange(0, 0),
        code="redis.call(\"lrange\", key_0, 0, 0)",
        teardown=lambda key: redis.delete(key),
        output_setup=3,
        output=["a"],
        output_teardown=1)

    check(
        "lrange_1",
        setup=lambda key: redis.rpush(key, "a", "b", "c"),
        normal=lambda key: redis.lrange(key, -3, 2),
        setup_pipe=lambda pipe, key: pipe.rpush(key, "a", "b", "c"),
        pipeline=lambda pipe, key: pipe.lrange(key, -3, 2),
        lua=lambda ctx, key: RedisList(key).lrange(-3, 2),
        code="redis.call(\"lrange\", key_0, -3, 2)",
        teardown=lambda key: redis.delete(key),
        output_setup=3,
        output=["a", "b", "c"],
        output_teardown=1)

    check(
        "lrange_2",
        setup=lambda key: redis.rpush(key, "a", "b", "c", "d", "e"),
        normal=lambda key: redis.lrange(key, 1, -2),
        setup_pipe=lambda pipe, key: pipe.rpush(key, "a", "b", "c", "d", "e"),
        pipeline=lambda pipe, key: pipe.lrange(key, 1, -2),
        lua=lambda ctx, key: RedisList(key).lrange(1, -2),
        code="redis.call(\"lrange\", key_0, 1, -2)",
        teardown=lambda key: redis.delete(key),
        output_setup=5,
        output=["b", "c", "d"],
        output_teardown=1)

    check(
        "lrange_3",
        setup=lambda key: redis.rpush(key, "a", "b", "c", "d", "e"),
        normal=lambda key: redis.lrange(key, -100, 100),
        setup_pipe=lambda pipe, key: pipe.rpush(key, "a", "b", "c", "d", "e"),
        pipeline=lambda pipe, key: pipe.lrange(key, -100, 100),
        lua=lambda ctx, key: RedisList(key).lrange(-100, 100),
        code="redis.call(\"lrange\", key_0, -100, 100)",
        teardown=lambda key: redis.delete(key),
        output_setup=5,
        output=["a", "b", "c", "d", "e"],
        output_teardown=1)
